import unittest
import asyncio
import signal
from zof.run_server import run_server


class RunServerTestCase(unittest.TestCase):
    def test_stopping_task(self):
        """Verify with task that calls stop.
        """
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        async def _task():
            asyncio.get_event_loop().stop()

        asyncio.ensure_future(_task())
        run_server()
        self.assert_final_task_count(0)

    def test_empty_task(self):
        """Verify with task that immediately returns.

        `run_server` will continue running even when there are no tasks
        scheduled at all. Use an alarm signal to force the test to finish.
        """
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        async def _task():
            pass

        asyncio.ensure_future(_task())
        signal.alarm(1)
        run_server(signals=['SIGALRM'])
        self.assert_final_task_count(0)

    def test_destroyed_task_pending(self):
        """Verify with task that sleeps in an infinite loop calling stop.
        """
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        async def _task():
            while True:
                await asyncio.sleep(0.1)
                asyncio.get_event_loop().stop()

        task = asyncio.ensure_future(_task())
        run_server(pending_timeout=0.001)
        self.assertTrue(loop.is_closed())

        tasks = asyncio.Task.all_tasks(loop)
        self.assertEqual({task}, tasks)
        with self.assertRaisesRegex(RuntimeError, 'Event loop is closed'):
            # This call may cause a 'Task was destroyed but it is pending!'
            # message. This is okay, but let's try to disable the log message.
            task._log_destroy_pending = False
            task.cancel()

    def assert_final_task_count(self, n):
        """Verify that event loop is closed and there are `n` tasks pending.
        """
        loop = asyncio.get_event_loop()
        self.assertTrue(loop.is_closed())
        self.assertEqual(n, len(asyncio.Task.all_tasks(loop)))



import asyncio
from zof.event import load_event


class Protocol(asyncio.SubprocessProtocol):
    """Implements an asyncio Protocol for parsing data received from oftr.
    """

    def __init__(self, post_event):
        self.post_event = post_event
        self.buf = b''
        self.exit_future = asyncio.Future()

    def pipe_data_received(self, fd, data):
        begin = 0
        offset = len(self.buf)
        self.buf += data
        while True:
            offset = self.buf.find(b'\x00', offset)
            if offset < 0:
                self.buf = self.buf[begin:]
                return
            if begin != offset:
                self.post_event(load_event(self.buf[begin:offset]))
            offset += 1
            begin = offset

    def process_exited(self):
        self.exit_future.set_result(0)



import inspect
import logging
import asyncio

LOGGER = logging.getLogger(__package__)

_MSG_SUBTYPES = {
    "CHANNEL_UP", "CHANNEL_DOWN", "CHANNEL_ALERT", "HELLO", "ERROR",
    "ECHO_REQUEST", "ECHO_REPLY", "EXPERIMENTER", "FEATURES_REQUEST",
    "FEATURES_REPLY", "GET_CONFIG_REQUEST", "GET_CONFIG_REPLY", "SET_CONFIG",
    "PACKET_IN", "FLOW_REMOVED", "PORT_STATUS", "PACKET_OUT", "FLOW_MOD",
    "GROUP_MOD", "PORT_MOD", "TABLE_MOD", "REQUEST.DESC", "REQUEST.FLOW",
    "REQUEST.AGGREGATE", "REQUEST.TABLE", "REQUEST.PORT_STATS",
    "REQUEST.QUEUE", "REQUEST.GROUP", "REQUEST.GROUP_DESC",
    "REQUEST.GROUP_FEATURES", "REQUEST.METER", "REQUEST.METER_CONFIG",
    "REQUEST.METER_FEATURES", "REQUEST.TABLE_FEATURES", "REQUEST.PORT_DESC",
    "REQUEST.TABLE_DESC", "REQUEST.QUEUE_DESC", "REQUEST.FLOW_MONITOR",
    "REPLY.DESC", "REPLY.FLOW", "REPLY.AGGREGATE", "REPLY.TABLE",
    "REPLY.PORT_STATS", "REPLY.QUEUE", "REPLY.GROUP", "REPLY.GROUP_DESC",
    "REPLY.GROUP_FEATURES", "REPLY.METER", "REPLY.METER_CONFIG",
    "REPLY.METER_FEATURES", "REPLY.TABLE_FEATURES", "REPLY.PORT_DESC",
    "REPLY.TABLE_DESC", "REPLY.QUEUE_DESC", "REPLY.FLOW_MONITOR",
    "BARRIER_REQUEST", "BARRIER_REPLY", "QUEUE_GET_CONFIG_REQUEST",
    "QUEUE_GET_CONFIG_REPLY", "ROLE_REQUEST", "ROLE_REPLY",
    "GET_ASYNC_REQUEST", "GET_ASYNC_REPLY", "SET_ASYNC", "METER_MOD",
    "ROLE_STATUS", "TABLE_STATUS", "REQUESTFORWARD", "BUNDLE_CONTROL",
    "BUNDLE_ADD_MESSAGE"
}


def make_handler(callback, type_, subtype='', options=None):
    """Factory function to create appropriate handler.

    Supported types: message, event
    """
    if type_ == 'message':
        return MessageHandler(callback, type_, subtype, options)
    if type_ == 'event':
        return EventHandler(callback, type_, subtype, options)
    raise ValueError('make_handler: Unknown handler type: "%s"' % type_)


class BaseHandler(object):
    """A Handler is a wrapper around an event callback.

    Attributes:
        callback (function): Callback function or coroutine
        type (str): "message", "event" or "command"
        subtype (str | function): Event or message subtype
        options (Optional[dict]): Handler options.
    """

    def __init__(self, callback, type_, subtype='', options=None):
        self.callback = callback
        self.type = type_
        self.subtype = subtype.upper() if isinstance(subtype, str) else subtype
        self.options = options

    def match(self, event):
        raise NotImplementedError("Please implement this method")

    def __call__(self, event, app):
        """Invoke handler and ignore the return value."""
        datapath_id = event.get('datapath_id')
        conn_id = event.get('conn_id')
        if asyncio.iscoroutinefunction(self.callback):
            app.ensure_future(
                self.callback(event), datapath_id=datapath_id, conn_id=conn_id)
        else:
            task = asyncio.Task.current_task()
            if task:
                task.zof_task_app = app
                task.zof_task_locals = {
                    'datapath_id': datapath_id,
                    'conn_id': conn_id
                }
            self.callback(event)

    def __repr__(self):
        cb_name = self.callback.__name__ if self.callback else 'NONE'
        return '%s[%s] %s %r' % (self.type, self.subtype, cb_name,
                                 self.options)

    def verify(self):
        return _verify_callback(self.callback, 1)

    def help(self):
        """Return help text."""
        text = self.callback.__doc__
        if not text:
            return 'No help available'
        return inspect.cleandoc(text)

    def help_brief(self):
        """Return summary line from help text."""
        text = self.callback.__doc__
        if not text:
            return 'No help available'
        return text.strip().split('\n', 1)[0]


class MessageHandler(BaseHandler):
    def match(self, event):
        # Check subtype to see if we can return False immediately.
        event_type = event['type']
        if callable(self.subtype):
            if not self.subtype(event_type):
                return False
        elif self.subtype != event_type:
            return False
        # Check for events that don't have a datapath_id. For these, the app
        # must explicitly opt in using `datapath_id=None`.
        if 'datapath_id' not in event:
            if self.options.get('datapath_id', '') is not None:
                return False
        else:
            if self.options.get('datapath_id', '') is None:
                return False
        # Check for matching option values in event.
        for key, value in self.options.items():
            if not _match_message(key, value, event):
                return False
        return True

    def verify(self):
        if not _verify_callback(self.callback, 1):
            return False
        if callable(self.subtype):
            if not _verify_callback(self.subtype, 1):
                LOGGER.warning('Message handler invalid subtype: %r',
                               self.subtype)
                return False
        elif self.subtype not in _MSG_SUBTYPES:
            LOGGER.warning(
                'Message handler subtype not recognized: %s',
                self.subtype,
                stack_info=True)
        return True


class EventHandler(BaseHandler):
    def match(self, event):
        # Check subtype to see if we can return false immediately.
        if callable(self.subtype):
            if not self.subtype(event['event']):
                return False
        elif event['event'] != self.subtype:
            return False
        # Check for matching option values in event.
        for key, value in self.options.items():
            if not _match_event(key, value, event):
                return False
        return True

    def verify(self):
        if not _verify_callback(self.callback, 1):
            return False
        if callable(self.subtype):
            if not _verify_callback(self.subtype, 1):
                LOGGER.warning('Event handler invalid subtype: %r',
                               self.subtype)
                return False
        return True


def _verify_callback(callback, param_count):
    """Make sure callback  has the expected number of positional parameters."""
    if not callable(callback):
        LOGGER.error('Callback is not a callable function: %s', callback)
        return False
    sig = inspect.signature(callback)
    if len(sig.parameters) != param_count:
        LOGGER.error('Callback has unexpected number of parameters: %s',
                     callback)
        return False
    return True


def _match_message(key, value, event):
    """Return true if `key` and `value` exist within the given message."""
    if key == 'datapath_id' and value is None:
        return True
    val = str(value).upper()
    msg = event['msg']
    if key in msg:
        return str(msg[key]).upper() == val
    pkt = msg.get('pkt')
    if pkt is not None and key in pkt:
        return str(pkt[key]).upper() == val
    return False


def _match_event(key, value, event):
    """Return true if `key` and `value` exist within the given event."""
    val = str(value).upper()
    if key in event:
        return str(event[key]).upper() == val
    return False



"""
Implements run() function.
"""

import sys
from .api_args import common_args
from .controller import Controller
from .logging import init_logging
from .pidfile import PidFile


def run(*, args=None):
    """Run event loop for zof.

    Args:
        args (Optional[argparse.Namespace]): Arguments derived from
            ArgumentParser. If None, use `common_args` parser. If args is a
            list, pass these to parse_args.
    """
    if args is None:
        args = common_args(include_x_modules=True).parse_args()
    elif isinstance(args, (list, tuple)):
        args = common_args(include_x_modules=True).parse_args(args)

    if args.loglevel:
        init_logging(args.loglevel, args.logfile)

    if args.xp_uvloop:
        # Replace default asyncio event loop with uvloop.
        import uvloop  # pylint: disable=import-error
        import asyncio
        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

    if args.xp_ujson:
        # Replace default json deserializer with ujson.
        import ujson  # pylint: disable=import-error
        import zof.objectview as _ov
        _ov.from_json = ujson.loads

    with PidFile(args.pidfile):
        controller = Controller.singleton()
        exit_status = controller.run_loop(args=args)

    if not args.x_under_test:
        sys.exit(exit_status)

    return exit_status



import unittest
import timeit
import pickle
from zof.objectview import from_json


def _test_performance(data):
    stmt = '''from_json("""%s""")''' % data
    result = timeit.timeit(
        stmt, setup='from zof.objectview import from_json', number=10000)
    print('test_from_json_speed=%r' % result)

    try:
        result = timeit.timeit(
            stmt, setup='from ujson import loads as from_json', number=10000)
        print('test_ujson_speed=%r' % result)
    except ModuleNotFoundError:
        pass

    obj = from_json(data)
    data = pickle.dumps(data)
    stmt = '''pickle.loads(%r)''' % data
    result = timeit.timeit(stmt, setup='import pickle', number=10000)
    print('test_pickle_speed=%r' % result)


class JsonTestCase(unittest.TestCase):
    def test_from_json(self):
        obj = from_json('{"b":2}')
        self.assertEqual(obj, {'b': 2})

        obj = from_json(b'{"a":1}')
        self.assertEqual(obj, {'a': 1})

    @unittest.skip("skip speed test")
    def test_from_json_speed(self):
        # Original PacketIn message
        data = '{"params":{"type":"PACKET_IN","time":"1506453847.618103000","xid":344,"version":1,"conn_id":3,"datapath_id":"00:00:00:00:00:00:00:01","msg":{"buffer_id":343,"total_len":64,"in_port":1,"in_phy_port":0,"metadata":0,"reason":"TABLE_MISS","table_id":0,"cookie":0,"match":[],"data":"8000000000010056010000010800450000320000000040FFF72CC0A80028C0A801287A18586B110897F519E2657E07CC31C311C7C40C8B955151335451D50036","_pkt":[{"field":"ETH_DST","value":"80:00:00:00:00:01"},{"field":"ETH_SRC","value":"00:56:01:00:00:01"},{"field":"ETH_TYPE","value":2048},{"field":"IP_DSCP","value":0},{"field":"IP_ECN","value":0},{"field":"IP_PROTO","value":255},{"field":"IPV4_SRC","value":"192.168.0.40"},{"field":"IPV4_DST","value":"192.168.1.40"},{"field":"NX_IP_TTL","value":64},{"field":"X_PKT_POS","value":34}]}},"method":"OFP.MESSAGE"}'
        _test_performance(data)

        # Smaller JSON
        data = '{"params":{"type":"PACKET_IN","time":"1506453847.618103000","xid":344,"version":1,"conn_id":3,"msg":{"buffer_id":343,"total_len":64,"in_port":1,"metadata":0,"reason":"TABLE_MISS","table_id":0,"cookie":0,"data":"8000000000010056010000010800450000320000000040FFF72CC0A80028C0A801287A18586B110897F519E2657E07CC31C311C7C40C8B955151335451D50036","_pkt":[{"field":"ETH_DST","value":"80:00:00:00:00:01"},{"field":"ETH_SRC","value":"00:56:01:00:00:01"},{"field":"ETH_TYPE","value":2048},{"field":"IP_PROTO","value":255},{"field":"IPV4_SRC","value":"192.168.0.40"},{"field":"IPV4_DST","value":"192.168.1.40"},{"field":"X_PKT_POS","value":34}]}},"method":"OFP.MESSAGE"}'
        _test_performance(data)

        # Smaller and Flatter JSON
        data = '{"params":{"type":"PACKET_IN","time":"1506453847.618103000","xid":344,"version":1,"conn_id":3,"msg":{"buffer_id":343,"total_len":64,"in_port":1,"metadata":0,"reason":"TABLE_MISS","table_id":0,"cookie":0,"data":"8000000000010056010000010800450000320000000040FFF72CC0A80028C0A801287A18586B110897F519E2657E07CC31C311C7C40C8B955151335451D50036","ETH_DST":"80:00:00:00:00:01","ETH_SRC":"00:56:01:00:00:01","ETH_TYPE":2048,"IP_PROTO":255,"IPV4_SRC":"192.168.0.40","IPV4_DST":"192.168.1.40","X_PKT_POS":34}},"method":"OFP.MESSAGE"}'
        _test_performance(data)



"""Implements ControllerApp class."""

import logging
import os
import signal
from operator import attrgetter
from .handler import make_handler
from . import exception as _exc


class ControllerApp(object):
    """Concrete class representing a controller application.

    Attributes:
        name (str): App name.
        ref (Application): Externally visible ref to this object.
        precedence (int): App precedence.
        arg_parser (ArgumentParser): Argument parser for this app.
        exception_fatal (bool|str): Terminate immediately if app raises
            exception.
        controller (Controller): App's parent controller object.
        logger (Logger): App's logger.
        handlers (Dict[str,List[BaseHandler]]): App handlers.
    Args:
        controller (Controller): Parent controller object.
        name (str): App name.
        ref (Application): Externally visible ref to this object.
        exception_fatal (bool): Terminate immediately if app raises exception.
        precedence (int): App precedence.
        arg_parser (ArgumentParser): Argument parser for this app.
        has_datapath_id (bool): If False, this app only handles messages
            without a datapath_id.
    """
    _curr_app_id = 0

    def __init__(self, *, controller, name, ref, exception_fatal, precedence,
                 arg_parser, has_datapath_id):
        self.name = name
        self.ref = ref
        self.precedence = precedence
        self.handlers = {}
        self.exception_fatal = exception_fatal
        self.arg_parser = arg_parser
        self._has_datapath_id = has_datapath_id
        self.set_controller(controller)

        self.logger = logging.getLogger('%s.%s' % (__package__, self.name))
        self.logger.info('Create app "%s"', self.name)

    def set_controller(self, controller):
        """Set controller parent."""
        self.controller = controller
        # Insert app into controller's list sorted by precedence.
        controller.apps.append(self)
        controller.apps.sort(key=attrgetter('precedence'), reverse=True)

    def handle_event(self, event, handler_type):
        """Handle event."""
        try:
            for handler in self.handlers.get(handler_type, []):
                if handler.match(event):
                    handler(event, self)
                    break
        except (_exc.StopPropagationException, _exc.PreflightUnloadException):
            # Pass this exception up to controller.
            raise
        except Exception:  # pylint: disable=broad-except
            self.handle_exception(event, handler_type)

    def handle_exception(self, event, handler_type):
        """Handle exception."""
        fatal_str = 'Fatal ' if self.exception_fatal else ''
        self.logger.critical(
            '%sException caught while handling "%s": %r',
            fatal_str,
            handler_type,
            event,
            exc_info=True)
        if self.exception_fatal:
            # If the value of `exception_fatal` is a string, find the logger
            # with this name and log the critical exception here also.
            if isinstance(self.exception_fatal, str):
                logger = logging.getLogger(self.exception_fatal)
                logger.critical(
                    '%sException caught while handling "%s": %r',
                    fatal_str,
                    handler_type,
                    event,
                    exc_info=True)
            # Shutdown the program immediately.
            logging.shutdown()
            os.kill(os.getpid(), signal.SIGKILL)

    def ensure_future(self, coroutine, *, datapath_id=None, conn_id=None):
        """Function used by an app to run an async handler.
        """
        return self.controller.ensure_future(
            coroutine, app=self, datapath_id=datapath_id, conn_id=conn_id)

    def register(self, callback, type_, subtype, options):
        """Function used to register a handler."""
        if self.controller.phase != 'INIT':
            raise RuntimeError('Controller not in INIT phase')
        if type_ == 'message' and not self._has_datapath_id and 'datapath_id' not in options:
            options['datapath_id'] = None
        handler = make_handler(callback, type_, subtype, options)
        if not handler.verify():
            self.logger.error('Failed to register %s', handler)
            return None
        handlers = self.handlers.setdefault(handler.type, [])
        handlers.append(handler)
        self.logger.debug('Register handler %s', handler)
        return callback



import os
import codecs
from ctypes import cdll, c_void_p, c_size_t, c_uint32, create_string_buffer
from zof.connection import Connection
from zof.objectview import to_json, ObjectView
from zof.exception import CodecError

OFTR_DLL = None

OFTR_ENCODE = 1
OFTR_DECODE = 2


def _dll():
    global OFTR_DLL  # pylint: disable=global-statement
    if OFTR_DLL is not None:
        return OFTR_DLL

    executable = Connection.find_oftr_path(os.getenv('ZOF_OFTR_PATH'))
    if not executable or executable[0] != '/':
        raise RuntimeError('Unable to locate oftr executable')

    OFTR_DLL = cdll.LoadLibrary(executable)
    OFTR_DLL.oftr_call.argtypes = [
        c_uint32, c_void_p, c_size_t, c_void_p, c_size_t
    ]
    return OFTR_DLL


def _oftr_call(opcode, data, buflen=2048):
    assert isinstance(data, bytes)

    dll = _dll()
    buf = create_string_buffer(buflen)

    result = dll.oftr_call(opcode, data, len(data), buf, buflen)
    if result < -buflen:
        # Result buffer is not big enough.
        buflen = -result
        buf = create_string_buffer(buflen)
        result = dll.oftr_call(opcode, data, len(data), buf, buflen)

    if result >= 0:
        # Buffer contains result.
        return buf[:result]

    if result >= -buflen:
        # Buffer contains error message.
        raise CodecError(buf[:-result].decode('utf-8'))

    # Result buffer is not big enough; this shouldn't happen.
    raise CodecError('_oftr_call failed: %r bytes needed' % result)


def encode(text, version=4):
    """Encode source code as binary.
    """

    if isinstance(text, (dict, ObjectView)):
        text = to_json(text)

    opcode = OFTR_ENCODE + (version << 24)
    return _oftr_call(
        opcode, text.encode('utf-8'), buflen=max(len(text), 1024))


def decode(binary):
    """Decode binary as source code.
    """

    return _oftr_call(OFTR_DECODE, bytes(binary), buflen=1024).decode('utf-8')


#-------------------------------------------------------------------------------


def _of_encode(text, errors='strict'):
    if errors != 'strict':
        raise CodecError('invalid errors argument: %s' % errors)
    return encode(text), len(text)


def _of_decode(binary, errors='strict'):
    if errors != 'strict':
        raise CodecError('invalid errors argument: %s' % errors)
    return decode(binary), len(binary)


def search(name):
    """Search function registered for codecs.

    Args:
        name (str): Codec name.

    Returns:
        CodecInfo: Encode/decode information or None if not found.
    """
    if name == 'openflow':
        return codecs.CodecInfo(
            name=name, encode=_of_encode, decode=_of_decode)
    return None


codecs.register(search)



import asyncio
from .controller import Controller
from .service.datapath import APP as DATAPATH_APP


def get_apps():
    """Get list of app's.
    """
    return [app.ref for app in Controller.singleton().apps]


def set_apps(apps):
    """Set list of apps.

    Avoid this function; it is provided for testing.
    """
    # pylint: disable=protected-access
    Controller.singleton().apps = [app._app for app in apps]


def get_datapaths():
    """Get list of currently connected datapaths.
    """
    return DATAPATH_APP.get_datapaths()


def find_datapath(*, datapath_id):
    """Return given datapath object.
    """
    return DATAPATH_APP.find_datapath(datapath_id)


def find_port(*, datapath_id, port_no):
    """Return given port object.
    """
    return DATAPATH_APP.find_port(datapath_id, port_no)


def post_event(event):
    """Function used to send an internal event to all app modules.

    Args:
        event (dict): event object
    """
    if not isinstance(event, dict):
        raise ValueError('%s not a dictionary: %r' % (type(event), event))
    Controller.singleton().post_event(event)


def ensure_future(coroutine, *, datapath_id=None, conn_id=None):
    """Function used by an app to run an async coroutine, under a specific
    scope.
    """
    app = asyncio.Task.current_task().zof_task_app
    return app.ensure_future(
        coroutine, datapath_id=datapath_id, conn_id=conn_id)


def _rpc_call(method, **params):
    """Function used to send a RPC request and receive a response.

    Returns:
        asyncio.Future: future for RPC reply
    """
    return Controller.singleton().rpc_call(method, **params)


async def connect(endpoint, *, options=(), versions=(), tls_id=0):
    """Make an outgoing OpenFlow connection.
    """
    result = await _rpc_call(
        'OFP.CONNECT',
        endpoint=endpoint,
        tls_id=tls_id,
        options=options,
        versions=versions)
    return result['conn_id']


async def close(*, conn_id=0, datapath_id=None):
    """Close an OpenFlow connection.
    """
    result = await _rpc_call(
        'OFP.CLOSE', conn_id=conn_id, datapath_id=datapath_id)
    return result['count']


async def get_connections(*, conn_id=0):
    """Get list of OpenFlow connections.
    """
    result = await _rpc_call('OFP.LIST_CONNECTIONS', conn_id=conn_id)
    return result['stats']


async def add_identity(*, cert, cacert, privkey):
    """Add a TLS identity.

    Returns:
        int: tls_id
    """
    result = await _rpc_call(
        'OFP.ADD_IDENTITY', cert=cert, cacert=cacert, privkey=privkey)
    return result['tls_id']



# N.B. This module must remain compatibile with Python 2.7.

try:
    # For Python 2.x compatibility.
    _BASESTRING = basestring
except NameError:
    _BASESTRING = str


def _convert_tp_dst(_key, ofctl):
    return '%s_dst' % _ip_proto_name(ofctl)


def _convert_tp_src(_key, ofctl):
    return '%s_src' % _ip_proto_name(ofctl)


def _convert_vlan(vlan):
    if isinstance(vlan, _BASESTRING):
        vlan = int(vlan, 0)
    if vlan > 0:
        vlan |= _OFPVID_PRESENT
    return vlan


_OFPVID_PRESENT = 0x1000
_ETHTYPE_ARP = 0x0806

_IP_PROTO_NAME = {6: 'tcp', 17: 'udp', 1: 'icmpv4', 58: 'icmpv6', 132: 'sctp'}

_LEGACY_FIELDS = dict(
    dl_type='eth_type',
    dl_src='eth_src',
    dl_dst='eth_dst',
    dl_vlan='vlan_vid',
    nw_src='ipv4_src',
    nw_dst='ipv4_dst',
    nw_proto='ip_proto',
    tp_dst=_convert_tp_dst,
    tp_src=_convert_tp_src)


def convert_from_ofctl(ofctl, validate=False):
    """Convert ofctl legacy field names."""
    if ofctl is None:
        return None
    result = {}
    for key, value in ofctl.items():
        key = convert_ofctl_field(key, ofctl)
        if key in result:
            raise ValueError('Duplicate ofctl field: %s' % key)
        result[key] = value
    if 'vlan_vid' in result:
        result['vlan_vid'] = _convert_vlan(result['vlan_vid'])
    if result.get('eth_type') == _ETHTYPE_ARP:
        _map_arp_fields(result)
    if validate:
        _validate_ofctl(result)
    return result


def convert_ofctl_field(key, ofctl):
    """Convert ofctl legacy field name."""
    conv = _LEGACY_FIELDS.get(key, key)
    return conv(key, ofctl) if callable(conv) else conv


def _ip_proto_name(ofctl):
    proto = ofctl.get('ip_proto') or ofctl.get('nw_proto')
    if proto is None:
        return 'tcp'
    try:
        return _IP_PROTO_NAME[proto]
    except KeyError:
        raise ValueError('Unknown ip_proto %s' % proto)


def _map_arp_fields(ofctl):
    """Translate nw_src/ipv4_src and nw_dst/ipv4_dst in the special case of
    an ARP match.

    At the point this function is called, nw_src/nw_dst has already been 
    renamed to ipv4_src/ipv4_dst.
    """
    if 'ipv4_src' in ofctl and 'arp_spa' not in ofctl:
        ofctl['arp_spa'] = ofctl.pop('ipv4_src')
    if 'ipv4_dst' in ofctl and 'arp_tpa' not in ofctl:
        ofctl['arp_tpa'] = ofctl.pop('ipv4_dst')


def _validate_ofctl(ofctl):
    """Validate the ofctl input.

    1. Each field has the appropriate type/syntax.
    2. There are no unrecognized fields.

    This function does not modify the input.
    """
    for key, value in ofctl.items():
        valid_fn = _VALID_FIELDS.get(key)
        if not valid_fn:
            raise ValueError('validate_ofctl: "%s" is not a valid field name' %
                             key)
        if not valid_fn(value):
            raise ValueError('validate_ofctl: field "%s: %s" is not valid' %
                             (key, value))


def _valid_port_no(value):
    return True


def _valid_int(value):
    try:
        int(str(value), 0)
    except ValueError:
        return False
    return True


def _valid_int_mask(value):
    if isinstance(value, str) and '/' in value:
        value, mask = value.split('/', 1)
        return _valid_int(value) and _valid_int(mask)
    return _valid_int(value)


def _valid_macaddr(value):
    return True


def _valid_ipv4(value):
    return True


def _valid_ipv6(value):
    return True


def _valid_vlan(value):
    return _valid_int_mask(value)


_VALID_FIELDS = {
    'in_port': _valid_port_no,
    'in_phy_port': _valid_int,
    'metadata': _valid_int_mask,
    'eth_dst': _valid_macaddr,
    'eth_src': _valid_macaddr,
    'eth_type': _valid_int,
    'vlan_vid': _valid_vlan,
    'vlan_pcp': _valid_int,
    'ip_dscp': _valid_int,
    'ip_ecn': _valid_int,
    'ip_proto': _valid_int,
    'ipv4_src': _valid_ipv4,
    'ipv4_dst': _valid_ipv4,
    'tcp_src': _valid_int_mask,
    'tcp_dst': _valid_int_mask,
    'udp_src': _valid_int_mask,
    'udp_dst': _valid_int_mask,
    'sctp_src': _valid_int_mask,
    'sctp_dst': _valid_int_mask,
    'icmpv4_type': _valid_int,
    'icmpv4_code': _valid_int,
    'arp_op': _valid_int,
    'arp_spa': _valid_ipv4,
    'arp_tpa': _valid_ipv4,
    'arp_sha': _valid_macaddr,
    'arp_tha': _valid_macaddr,
    'ipv6_src': _valid_ipv6,
    'ipv6_dst': _valid_ipv6,
    'ipv6_flabel': _valid_int,
    'icmpv6_type': _valid_int,
    'icmpv6_code': _valid_int,
    'ipv6_nd_target': _valid_ipv6,
    'ipv6_nd_sll': _valid_macaddr,
    'ipv6_nd_tll': _valid_macaddr,
    'mpls_label': _valid_int,
    'mpls_tc': _valid_int,
    'mpls_bos': _valid_int,
    'pbb_isid': _valid_int_mask,
    'tunnel_id': _valid_int_mask,
    'ipv6_exthdr': _valid_int_mask,
    'pbb_uca': _valid_int,
}



import asyncio
import zof

APP = zof.Application('logger', precedence=0)

APP.logger.info('phase = %r', APP.phase)


@APP.event('PREFLIGHT')
def preflight(event):
    APP.logger.info('event = %r', event)


@APP.event('POSTFLIGHT')
def postflight(event):
    APP.logger.info('event = %r', event)


@APP.event('SIGNAL')
def signal(event):
    APP.logger.info('event = %r', event)


@APP.event('PRESTART')
async def prestart(event):
    await logger(event)


@APP.event('START')
async def start(event):
    await logger(event)


@APP.event('STOP')
async def stop(event):
    await logger(event)


@APP.message(any)
@APP.event(any)
def other(event):
    APP.logger.warning('event = %r', event)


async def logger(event):
    try:
        APP.logger.info('async event = %r', event)
        await asyncio.sleep(3)
    except asyncio.CancelledError:
        APP.logger.warning('async event cancelled = %r', event)
    finally:
        APP.logger.info('async event done = %r', event)


if __name__ == '__main__':
    zof.run()



from collections import OrderedDict
from .controller import Controller

# Keys for datapath.user_data dictionary. Used by datapath service.
CHANNEL_UP_MSG = '_datapath.channel_up'
FEATURES_MSG = '_datapath.features_reply'


class DatapathList:
    """Represents a collection of datapaths."""

    def __init__(self):
        self._datapaths = OrderedDict()

    def add_datapath(self, *, datapath_id, conn_id):
        """Add datapath to list.

        Implementation is idempotent when arguments are identical.

        Returns:
            Datapath: datapath object
        """
        dpid_key = normalize_datapath_id(datapath_id)
        datapath = self._datapaths.get(dpid_key)
        if datapath is None:
            datapath = Datapath(datapath_id, conn_id)
            self._datapaths[dpid_key] = datapath
        elif datapath.conn_id != conn_id:
            raise ValueError('Datapath already exists: %r' % datapath)
        return datapath

    def delete_datapath(self, *, datapath_id):
        """Remove datapath from list.

        Returns None if datapath is not found.

        Returns:
            Datapath: just removed datapath object (or None)
        """
        dpid_key = normalize_datapath_id(datapath_id)
        datapath = self._datapaths.pop(dpid_key, None)
        return datapath

    def __len__(self):
        return len(self._datapaths)

    def __iter__(self):
        return iter(self._datapaths.values())

    def __getitem__(self, datapath_id):
        dpid_key = normalize_datapath_id(datapath_id)
        return self._datapaths[dpid_key]


class Datapath:
    """Represents a datapath."""

    def __init__(self, datapath_id, conn_id):
        self.datapath_id = datapath_id
        self.conn_id = conn_id
        self.ports = OrderedDict()
        self.up = True
        self.ready = False
        self.closed = False
        self.user_data = {}

    def add_port(self, *, port_no):
        """Add port to datapath.

        Implementation is idempotent when arguments are identical.

        Returns:
            Port: port object
        """
        port_no = normalize_port_no(port_no)
        port = self.ports.get(port_no)
        if port is None:
            port = Port(port_no, datapath=self)
            self.ports[port_no] = port
        return port

    def delete_port(self, *, port_no):
        """Remove port from datapath.

        Returns None if port is not found.

        Returns:
            Port: just removed port object (or None)
        """
        port_no = normalize_port_no(port_no)
        port = self.ports.pop(port_no, None)
        return port

    def add_ports(self, port_descs):
        """Add ports from OpenFlow Port descs.

        This is an idempotent operation; it can be used to update existing
        ports from a port_status message.

        Arguments:
            port_descs (List[ObjectView]): list of OpenFlow port desc's.
        """
        for port_desc in port_descs:
            port = self.add_port(port_no=port_desc['port_no'])
            port.hw_addr = port_desc['hw_addr']
            port.name = port_desc['name']
            port.state = port_desc['state']
            port.config = port_desc['config']

    def close(self):
        """Close connection to datapath; i.e. hang up.

        This function coordinates with the datapath service to make sure no
        apps receive any further events from this datapath.
        """
        if not self.closed:
            self.closed = True
            Controller.singleton().rpc_call(
                'OFP.CLOSE', ignore_result=True, conn_id=self.conn_id)

    def __getstate__(self):
        return str(self)

    def __len__(self):
        return len(self.ports)

    def __bool__(self):
        # Make sure empty datapath is still true.
        return True

    def __iter__(self):
        return iter(self.ports.values())

    def __getitem__(self, port_no):
        port_no = normalize_port_no(port_no)
        return self.ports[port_no]

    def __repr__(self):
        # Include single quotes; repr(self) should be valid YAML scalar.
        return "'<zof.Datapath %s>'" % self.datapath_id

    def __str__(self):
        return "<zof.Datapath %s>" % self.datapath_id


class Port:
    """Represents a datapath port."""

    def __init__(self, port_no, datapath):
        self.datapath = datapath
        self.port_no = port_no
        self.hw_addr = None
        self.name = None
        self.state = []
        self.config = []

    def __getstate__(self):
        return str(self)

    @property
    def datapath_id(self):
        return self.datapath.datapath_id

    @property
    def up(self):
        "Return true if port is up."
        return 'LINK_DOWN' not in self.state

    @property
    def admin_down(self):
        "Return true if port is administratively configured down."
        return 'PORT_DOWN' in self.config

    def __repr__(self):
        # Include single quotes; repr(self) should be valid YAML scalar.
        return "'<zof.Port %s>'" % self.port_no

    def __str__(self):
        return "<zof.Port %s>" % self.port_no


def normalize_datapath_id(datapath_id):
    """Normalize datapath_id value."""
    if isinstance(datapath_id, int):
        return datapath_id
    if isinstance(datapath_id, str) and datapath_id:
        return int(datapath_id.replace(':', ''), 16)
    raise ValueError('Invalid datapath_id: %r' % datapath_id)


def normalize_port_no(port_no):
    """Normalize port number value to int or string."""
    if isinstance(port_no, int):
        return port_no
    if isinstance(port_no, str):
        try:
            # Convert decimal and hexadecimal port number strings.
            return int(port_no, 0)
        except ValueError:
            return port_no.upper()
    raise ValueError('Invalid port_no: %r' % port_no)



import zof
from ..http import HttpServer
from ..pktview import pktview_from_list

APP = zof.Application('rest_api')
APP.http_endpoint = '127.0.0.1:8080'

WEB = HttpServer(logger=APP.logger)
# WEB.define_var('dpid', _parse_dpid)


@APP.event('start')
async def start(_):
    await WEB.start(APP.http_endpoint)


@APP.event('stop')
async def stop(_):
    await WEB.stop()


@WEB.get_json('/stats/switches')
def get_switches():
    return [d.datapath_id for d in zof.get_datapaths()]


@WEB.get_json('/stats/flow/{dpid:[0-9A-F:]+}')
async def get_flows(dpid):
    result = []
    async for ofmsg in FLOW_REQ.request(datapath_id=_parse_dpid(dpid)):
        _translate_flows(ofmsg['msg'])
        result.extend(ofmsg['msg'])
    return {dpid: result}


@WEB.get_json('/stats/groupdesc/{dpid:[0-9A-F:]+}')
async def get_groupdesc(dpid):
    result = await GROUPDESC_REQ.request(datapath_id=_parse_dpid(dpid))
    _translate_groups(result['msg'])
    return {dpid: result['msg']}


@WEB.get_json('/stats/port/{dpid}/{port_no}')
async def get_portstats(dpid, port_no):
    result = await PORTSTATS_REQ.request(
        datapath_id=_parse_dpid(dpid), port_no=_parse_port(port_no))
    return {dpid: result['msg']}


@WEB.get_json('/stats/port/{dpid}')
async def get_portstats(dpid):
    result = await PORTSTATS_REQ.request(
        datapath_id=_parse_dpid(dpid), port_no='ANY')
    return {dpid: result['msg']}


@WEB.post_json('/stats/portdesc/modify')
async def modify_portdesc(post_data):
    dpid = _parse_dpid(post_data['dpid'])
    port_no = _parse_port(post_data['port_no'])
    port = zof.find_port(datapath_id=dpid, port_no=port_no)
    PORTMOD_REQ.send(
        datapath_id=dpid,
        port_no=port_no,
        hw_addr=port.hw_addr,
        config=post_data['config'],
        mask=post_data['mask'])
    # FIXME(bfish): This code does not handle OpenFlow errors elicited from the PortMod
    # message. Any errors returned will only show up in the log. The barrier here is
    # just a cheap trick to verify that the portmod *should* have been acted on.
    result = await BARRIER_REQ.request(datapath_id=dpid)
    return result['msg']


@WEB.get_json('/stats/portdesc/{dpid}')
async def get_portdesc(dpid):
    result = await PORTDESC_REQ.request(datapath_id=_parse_dpid(dpid))
    return {dpid: result['msg']}


FLOW_REQ = zof.compile('''
    type: REQUEST.FLOW
    msg:
        table_id: ALL
        out_port: ANY
        out_group: ANY
        cookie: 0
        cookie_mask: 0
        match: []
''')

GROUPDESC_REQ = zof.compile('''
    type: REQUEST.GROUP_DESC
''')

PORTSTATS_REQ = zof.compile('''
    type: REQUEST.PORT_STATS
    msg:
        port_no: $port_no
''')

PORTDESC_REQ = zof.compile('''
    type: REQUEST.PORT_DESC
''')

PORTMOD_REQ = zof.compile('''
    type: PORT_MOD
    msg:
        port_no: $port_no
        hw_addr: $hw_addr
        config: [$config]
        mask: [$mask]
''')

BARRIER_REQ = zof.compile('''
    type: BARRIER_REQUEST
''')


def _parse_dpid(dpid):
    if isinstance(dpid, int):
        return _convert_dpid(dpid)
    if ':' in dpid:
        return dpid
    return _convert_dpid(int(dpid, 0))


def _convert_dpid(dpid):
    hexstr = '%16.16x' % dpid
    return ':'.join(hexstr[2 * i:2 * i + 2] for i in range(8))


def _parse_port(port_no):
    if isinstance(port_no, int):
        return port_no
    return int(port_no, 0)


def _translate_flows(msgs):
    for msg in msgs:
        if 'match' in msg:
            msg['match'] = pktview_from_list(msg['match'], slash_notation=True)
        if 'instructions' in msg:
            msg['actions'] = _translate_instructions(msg['instructions'])


def _translate_groups(msgs):
    for msg in msgs:
        for bkt in msg['buckets']:
            if 'actions' in bkt:
                bkt['actions'] = _translate_actions(bkt['actions'])


def _translate_instructions(instrs):
    result = []
    for instr in instrs:
        result += _translate_instruction(instr)
    return result


def _translate_instruction(instr):
    if instr['instruction'] == 'APPLY_ACTIONS':
        return _translate_actions(instr['actions'])
    return [str(instr)]


def _translate_actions(actions):
    return [_translate_action(act) for act in actions]


def _translate_action(action):
    action_type = action['action']
    if action_type == 'OUTPUT':
        return 'OUTPUT:%s' % action['port_no']
    if action_type == 'GROUP':
        return 'GROUP:%s' % action['group_id']
    if action_type == 'SET_FIELD':
        return 'SET_FIELD: {%s:%s}' % (action['field'].lower(),
                                       action['value'])
    if len(action) == 1:
        return '%s' % action_type
    return str(action)


if __name__ == '__main__':
    zof.run()



import unittest
import os
import tempfile
from zof.pidfile import PidFile


def _tmp_path():
    path = tempfile.gettempdir()
    return os.path.join(path, 'zof-pidfile-test')


class PidFileTestCase(unittest.TestCase):
    def test_pidfile(self):
        path = _tmp_path()
        pidfile = PidFile(path)
        pidfile.write()

        self.assertEqual(pidfile.read(), os.getpid())

        pidfile.remove()
        # Test that it's removed.
        self.assertFalse(os.path.exists(path))

    def test_context_mgr(self):
        path = _tmp_path()
        with PidFile(path):
            pid = PidFile(path).read()
            self.assertEqual(pid, os.getpid())



"""
This app retrieves a datapath's ports.

This is a guard app with high precedence. This app blocks other apps from
receiving channel_up messages until all of the datapath's ports are discovered.

This app modifies message events to include the source `datapath` object.
"""

import zof
from zof.datapath import DatapathList, CHANNEL_UP_MSG, FEATURES_MSG
import zof.exception as _exc


class DatapathApp(zof.Application):
    def __init__(self):
        super().__init__('service.datapath', precedence=1000000000)
        self.datapaths = DatapathList()

    def get_datapaths(self):
        return [datapath for datapath in self.datapaths if datapath.ready]

    def find_datapath(self, datapath_id):
        try:
            datapath = self.datapaths[datapath_id]
            if datapath.ready:
                return datapath
        except KeyError:
            pass
        return None

    def find_port(self, datapath_id, port_no):
        datapath = self.find_datapath(datapath_id)
        try:
            return datapath[port_no]
        except KeyError:
            return None


APP = DatapathApp()

PORTS_REQUEST = zof.compile('type: REQUEST.PORT_DESC')


@APP.message('channel_up')
def channel_up(event):
    datapath = APP.datapaths.add_datapath(
        datapath_id=event['datapath_id'], conn_id=event['conn_id'])
    if datapath.ready:
        event['datapath'] = datapath
        return
    datapath.user_data[CHANNEL_UP_MSG] = event
    raise _exc.StopPropagationException()


@APP.message('channel_down')
def channel_down(event):
    datapath = APP.datapaths.delete_datapath(datapath_id=event['datapath_id'])
    datapath.up = False
    if datapath.ready:
        event['datapath'] = datapath
        return
    raise _exc.StopPropagationException()


@APP.message('features_reply')
def features_reply(event):
    datapath = APP.datapaths[event['datapath_id']]
    if datapath.ready:
        event['datapath'] = datapath
        return

    datapath.user_data[FEATURES_MSG] = event
    if event['version'] > 1:
        zof.ensure_future(
            _get_ports(datapath), datapath_id=event['datapath_id'])
    else:
        datapath.add_ports(event['msg']['ports'])
        _post_channel_up(datapath)

    raise _exc.StopPropagationException()


@APP.message('port_status')
def port_status(event):
    datapath = APP.find_datapath(event['datapath_id'])
    if datapath is not None:
        event['datapath'] = datapath
        msg = event['msg']
        reason = msg['reason']
        if reason == 'DELETE':
            datapath.delete_port(port_no=msg['port_no'])
        elif reason in ('ADD', 'MODIFY'):
            datapath.add_ports([msg])
        else:
            APP.logger.warning('Unknown port_status reason: %r', event)
        return

    raise _exc.StopPropagationException()


@APP.message(any)
def other_message(event):
    datapath = APP.find_datapath(event['datapath_id'])
    if datapath is not None:
        event['datapath'] = datapath
        return

    raise _exc.StopPropagationException()


async def _get_ports(datapath):
    try:
        ports = await PORTS_REQUEST.request(
            datapath_id=datapath.datapath_id, conn_id=datapath.conn_id)
        datapath.add_ports(ports['msg'])
        _post_channel_up(datapath)
    except _exc.ControllerException as ex:
        datapath.close()
        APP.logger.warning('_get_ports failed: %s', ex)


def _post_channel_up(datapath):
    datapath.ready = True

    channel_event = datapath.user_data.pop(CHANNEL_UP_MSG)
    channel_event['event'] = 'MESSAGE'
    zof.post_event(channel_event)

    features = datapath.user_data.pop(FEATURES_MSG)
    features['event'] = 'MESSAGE'
    zof.post_event(features)



import string
import textwrap
import asyncio
import logging
import zof
from .controller import Controller
from .objectview import ObjectView, to_json, to_json_pretty
from .pktview import pktview_to_list
from .asyncmap import asyncmap

LOGGER = logging.getLogger(__package__)

_TEMPLATE = """\
method: OFP.SEND
params:
  xid: $xid
  datapath_id: $datapath_id
  conn_id: $conn_id
  %s"""


# pylint: disable=redefined-builtin
def compile(msg):
    """Compile an OpenFlow message template."""
    controller = Controller.singleton()
    if isinstance(msg, str):
        return CompiledString(controller, msg)
    if 'type' in msg:
        return CompiledObject(controller, msg)
    return CompiledObjectRPC(controller, msg)


class CompiledMessage:
    """Abstract class representing a compiled OpenFlow message.

    Attributes:
        _controller (Controller): Controller object.
    """

    _controller = None

    def send(self, **kwds):
        """Send an OpenFlow message (fire and forget).

        Args:
            kwds (dict): Template argument values.
        """
        kwds.setdefault('xid', self._controller.next_xid())
        self._controller.write(self._complete(kwds, _task_locals()))

    def request(self, **kwds):
        """Send an OpenFlow request and receive a response.

        Args:
            kwds (dict): Template argument values.
        """
        xid = kwds.setdefault('xid', self._controller.next_xid())
        return self._controller.write(
            self._complete(kwds, _task_locals()), xid)

    def request_all(self, *, parallelism=1, **kwds):
        """Send multiple OpenFlow requests and receive responses.

        Args:
            kwds (dict): Template argument values.
        """

        def _req(conn_id):
            return self.request(conn_id=conn_id, **kwds)

        conn_ids = [dp.conn_id for dp in zof.get_datapaths()]
        return asyncmap(_req, conn_ids, parallelism=parallelism)

    def _complete(self, kwds, task_locals):
        raise NotImplementedError()


class CompiledString(CompiledMessage):
    """Concrete class representing a compiled OpenFlow message template.

    Attributes:
        _controller (Controller): Controller object.
        _template (StringTemplate): Prepared message template.
    """

    def __init__(self, controller, msg):
        assert isinstance(msg, str)
        self._controller = controller
        self._template = None
        self._template_args = None
        self._compile(msg)

    def _compile(self, msg):
        """Compile OFP.SEND message and store it into `self._template`.

        Args:
            msg (str): YAML message.
        """
        # Remove top-level indent.
        msg = textwrap.dedent(msg).strip()
        # Add indent of 2 spaces.
        msg = msg.replace('\n', '\n  ')
        self._template = MyTemplate(_TEMPLATE % msg)
        self._template_args = self._template.args()

    def _complete(self, kwds, task_locals):
        """Substitute keywords into OFP.SEND template.

        Translate `bytes` values to hexadecimal and escape all string values.
        """

        dpid = kwds.setdefault('datapath_id', task_locals.get('datapath_id'))
        conn_id = kwds.setdefault('conn_id', task_locals.get('conn_id'))

        if conn_id is None:
            # Either conn_id is not present *or* it's equal to None.
            kwds['conn_id'] = 0
            if not dpid:
                raise ValueError('Must specify either datapath_id or conn_id.')

        for key in kwds:
            val = kwds[key]
            if isinstance(val, bytes):
                kwds[key] = val.hex()
            elif isinstance(val, (str, dict, ObjectView)):
                kwds[key] = to_json(val)
            elif val is None:
                kwds[key] = 'null'

        try:
            self._check(kwds)
            return self._template.substitute(kwds)
        except KeyError as ex:
            error = 'Missing ${%s} argument for %r' % (ex.args[0], self)
        raise LookupError(error)

    def _check(self, kwds):
        """Check kwds against known template vars."""
        for key in kwds:
            if key not in self._template_args:
                raise ValueError('Unknown keyword argument "%s"' % key)

    def __repr__(self):
        """String representation (used for testing)
        """
        return '<zof.CompiledString args=%r>\n%s\n</zof.CompiledString>' % (sorted(self._template_args), self._template.template)


class CompiledObject(CompiledMessage):
    """Concrete class representing a compiled OpenFlow object template."""

    def __init__(self, controller, obj):
        assert isinstance(obj, (dict, ObjectView))
        assert 'type' in obj
        self._controller = controller
        self._obj = obj
        if self._obj['type'] in ('PACKET_OUT', 'PACKET_IN'):
            self._convert_pkt()

    def _complete(self, kwds, task_locals):
        """Substitute keywords into object template, and compile to JSON.
        """
        kwds.setdefault('datapath_id', task_locals['datapath_id'])
        kwds.setdefault('conn_id', task_locals['conn_id'])

        if 'datapath_id' not in self._obj:
            self._obj['datapath_id'] = kwds['datapath_id']
        if 'xid' not in self._obj:
            self._obj['xid'] = kwds['xid']
        if 'conn_id' not in self._obj and kwds['conn_id'] is not None:
            self._obj['conn_id'] = kwds['conn_id']

        if self._obj['datapath_id'] is None and self._obj.get(
                'conn_id') is None:
            raise ValueError('Must specify either datapath_id or conn_id.')

        return to_json({'method': 'OFP.SEND', 'params': self._obj})

    def _convert_pkt(self):
        """Convert high level API `pkt` to low level API."""
        msg = self._obj['msg']
        if 'pkt' in msg:
            msg = msg.copy()
            if 'payload' in msg['pkt']:
                msg['_pkt_data'] = msg['pkt']['payload']
            msg['_pkt'] = pktview_to_list(msg['pkt'])
            del msg['pkt']
            self._obj['msg'] = msg

    def __repr__(self):
        """String representation (used for testing)
        """
        return '<zof.CompiledObject>\n%s\n</zof.CompiledObject>' % to_json_pretty(self._obj)


class CompiledObjectRPC(CompiledMessage):
    """Concrete class representing a compiled RPC message."""

    def __init__(self, controller, obj):
        assert isinstance(obj, (dict, ObjectView))
        assert 'method' in obj
        self._controller = controller
        self._obj = obj

    def send(self, **kwds):
        """Send an OpenFlow message (fire and forget).

        Args:
            kwds (dict): Template argument values.
        """
        self._controller.write(self._complete(kwds, _task_locals()))

    def _complete(self, kwds, task_locals):
        """Substitute keywords into object template, and compile to JSON.
        """
        return to_json(self._obj)

    def __repr__(self):
        """String representation (used for testing)
        """
        return '<zof.CompiledObjectRPC>\n%s\n</zof.CompiledObjectRPC>' % to_json_pretty(self._obj)


class MyTemplate(string.Template):
    def args(self):
        """Return set of template arguments."""
        result = set()
        # pylint: disable=no-member
        for m in self.pattern.finditer(self.template):
            named = m.group('named') or m.group('braced')
            if named:
                result.add(named)
        return result


def _task_locals():
    task = asyncio.Task.current_task()
    task_locals = getattr(task, 'zof_task_locals', {})
    assert isinstance(task_locals, dict)
    return task_locals



import argparse
import zof
from zof import exception as _exc
from zof.http import HttpServer
from prometheus_client import REGISTRY, CollectorRegistry, generate_latest, ProcessCollector
from prometheus_client.core import CounterMetricFamily, GaugeMetricFamily


def arg_parser():
    parser = argparse.ArgumentParser(
        prog='Metrics', description='Metric Demo', add_help=False)
    parser.add_argument(
        '--metrics-endpoint', help='HTTP endpoint for metrics server')
    return parser


APP = zof.Application('metrics', arg_parser=arg_parser())
WEB = HttpServer()


@APP.event('preflight')
def preflight(_):
    if not APP.args.metrics_endpoint:
        # If we're not listening, unload the application.
        raise _exc.PreflightUnloadException()


@APP.event('prestart')
async def start(_):
    # Start a process collector for our oftr subprocess.
    ProcessCollector(namespace='oftr', pid=lambda: APP.oftr_connection.pid)
    await WEB.start(APP.args.metrics_endpoint)
    APP.logger.info('Start listening on %s', APP.args.metrics_endpoint)


@APP.event('stop')
async def stop(_):
    await WEB.stop()
    APP.logger.info('Stop listening on %s', APP.args.metrics_endpoint)


@WEB.get_text('/')
@WEB.get_text('/metrics')
@WEB.get_text('/metrics/')
async def metrics():
    return generate_latest(REGISTRY)


@WEB.get_text('/metrics/ports?{target}')
async def ports(target):
    if target:
        met = PortMetrics()
        await _collect_port_stats(target, met)
    else:
        met = PortMetrics(include_instance=True)
        await _collect_port_stats_all(met)
    return _dump_prometheus(met.metrics())


PORT_STATS = zof.compile('''
type: REQUEST.PORT_STATS
msg:
  port_no: ANY
''')


def _supported_counter(value):
    return value != 0xffffffffffffffff


class PortMetrics:
    def __init__(self, include_instance=False):
        self.include_instance = include_instance
        if include_instance:
            labels = ['port_no', 'instance']
        else:
            labels = ['port_no']
        self.tx_bytes = CounterMetricFamily('port_tx_bytes_total',
                                            'bytes transmitted', None, labels)
        self.rx_bytes = CounterMetricFamily('port_rx_bytes_total',
                                            'bytes received', None, labels)
        self.tx_packets = CounterMetricFamily(
            'port_tx_packets_total', 'packets transmitted', None, labels)
        self.rx_packets = CounterMetricFamily('port_rx_packets_total',
                                              'packets received', None, labels)
        self.tx_dropped = CounterMetricFamily(
            'port_tx_drops_total', 'packets dropped by TX', None, labels)
        self.rx_dropped = CounterMetricFamily(
            'port_rx_drops_total', 'packets dropped by RX', None, labels)
        self.rx_errors = CounterMetricFamily('port_rx_errors_total',
                                             'receive errors', None, labels)
        self.duration = CounterMetricFamily(
            'port_duration_seconds_total', 'duration in seconds', None, labels)
        self.port_up = GaugeMetricFamily('port_up', 'port is up', None, labels)

    def metrics(self):
        return [
            self.tx_bytes, self.rx_bytes, self.tx_packets, self.rx_packets,
            self.tx_dropped, self.rx_dropped, self.rx_errors, self.duration,
            self.port_up
        ]

    def update(self, dpid, stat):
        port_no = stat['port_no']
        if self.include_instance:
            labels = [str(port_no), dpid]
        else:
            labels = [str(port_no)]
        for counter, value in [(self.tx_bytes, stat['tx_bytes']),
                               (self.rx_bytes, stat['rx_bytes']),
                               (self.tx_packets, stat['tx_packets']),
                               (self.rx_packets, stat['rx_packets']),
                               (self.tx_dropped, stat['tx_dropped']),
                               (self.rx_dropped, stat['rx_dropped']),
                               (self.rx_errors, stat['rx_errors'])]:
            if _supported_counter(value):
                counter.add_metric(labels, value)
        if stat['duration'] != '0':
            self.duration.add_metric(labels, float(stat['duration']))
        port = zof.find_port(datapath_id=dpid, port_no=port_no)
        if port:
            self.port_up.add_metric(labels, int(port.up))


async def _collect_port_stats(dpid, metric):
    try:
        reply = await PORT_STATS.request(datapath_id=dpid)
        for stat in reply['msg']:
            metric.update(reply['datapath_id'], stat)
    except _exc.ControllerException as ex:
        APP.logger.warning('Unable to retrieve stats for dpid %s: %r', dpid,
                           ex)


async def _collect_port_stats_all(metric):
    async for reply in PORT_STATS.request_all(parallelism=5):
        try:
            reply = await reply
            for stat in reply['msg']:
                metric.update(reply['datapath_id'], stat)
        except _exc.ControllerException:
            pass


class _MyCollector:
    def __init__(self, stats):
        self.stats = stats

    def collect(self):
        return self.stats


def _dump_prometheus(stats):
    registry = CollectorRegistry()
    registry.register(_MyCollector(stats))
    return generate_latest(registry)


if __name__ == '__main__':
    zof.run()



"""Implements Controller class."""

import asyncio
import logging
import sys
import functools
import inspect
from collections import defaultdict
from .event import load_event, dump_event
from .pktview import pktview_from_list
from .connection import Connection
from .run_server import run_server
from . import exception as _exc

_XID_TIMEOUT = 10.0  # Seconds
_IDLE_INTERVAL = 1.0
_MIN_XID = 10000
_MAX_XID = 0xFFFFFFFF
_API_VERSION = 0.9
_VERSION = '0.1.0'

LOGGER = logging.getLogger(__package__)


class Controller(object):
    """Concrete class that supports multiple app modules.

    Attributes:
        apps (List[ControllerApp]): List of apps ordered by precedence.
        args (argparse.Namespace): Arguments parsed by argparse module.
        phase (str): Lifecycle phase.
        conn (Connection): oftr connection.
    """

    _singleton = None

    @staticmethod
    def singleton():
        """Return global singleton object."""
        if not Controller._singleton:
            Controller._singleton = Controller()
        return Controller._singleton

    def __init__(self):
        self.apps = []
        self.args = None
        self.phase = 'INIT'
        self.conn = None
        self._xid = _MIN_XID
        self._reqs = {}
        self._event_queue = None
        self._supported_versions = []
        self._tls_id = 0
        self._tasks = defaultdict(list)
        self._exit_status = 1

    def find_app(self, name):
        """Find application object by name."""
        return any(app for app in self.apps if app.name == name)

    def run_loop(self, *, args):
        """Main entry point for running a controller.

        Returns exit status.
        """
        if not self.apps:
            LOGGER.warning('No apps are loaded.')

        self.args = args

        try:
            asyncio.ensure_future(self._run())
            run_server(
                signals=['SIGTERM', 'SIGINT', 'SIGHUP'],
                signal_handler=self._handle_signal,
                logger=LOGGER)
            self._set_phase('POSTFLIGHT')

        except Exception as ex:  # pylint: disable=broad-except
            LOGGER.exception(ex)

        finally:
            LOGGER.info('Exiting with status %d', self._exit_status)

        return self._exit_status

    def _handle_signal(self, signame):
        """Handle signals.

        Usually, signals indicate that the program should exit. An app can
        prevent shutdown by setting the event's `exit` value to False.
        """
        LOGGER.info('Signal Received: %s (qsize=%d)', signame,
                    self._event_queue.qsize())
        self.post_event({'event': 'SIGNAL', 'signal': signame, 'exit': True})

    async def _run(self):
        """Async task for running the controller."""
        LOGGER.debug("Controller._run entered")
        try:
            self._preflight()

            self.conn = Connection(oftr_options={
                'path': self.args.x_oftr_path,
                'args': self.args.x_oftr_args,
                'prefix': self.args.x_oftr_prefix
            })
            # Call connect() with "self.post_event" to use protocol based api.
            proto_callback = self.post_event if not self.args.xp_streams else None
            await self.conn.connect(proto_callback)

            self._event_queue = asyncio.Queue()
            self._set_phase('PRESTART')

            asyncio.ensure_future(self._start())
            if not proto_callback:
                # Schedule the read loop to read from the stream, if we're not 
                # using the protocol api.
                asyncio.ensure_future(self._read_loop())
            idle = asyncio.ensure_future(self._idle_task())

            await self._event_loop()

            idle.cancel()
            self._set_phase('STOP')
            await self.conn.disconnect()

        except Exception:  # pylint: disable=broad-except
            LOGGER.exception('Exception in Controller._run')
            if self.conn:
                self.conn.close(False)
            asyncio.get_event_loop().stop()

        finally:
            LOGGER.debug("Controller._run exited")

    async def _event_loop(self):
        """Run the event loop to handle events."""
        try:
            LOGGER.debug('_event_loop entered')
            while True:
                event = await self._event_queue.get()
                self._dispatch_event(event)
                # If the event was dispatched to an async task, give it time
                # now to run so it can get started.
                await asyncio.sleep(0)
        except _exc.ExitException as ex:
            self._exit_status = ex.exit_status
            self.conn.close(True)
            asyncio.get_event_loop().stop()
        except Exception:  # pylint: disable=broad-except
            LOGGER.exception('Exception in Controller._event_loop')
            sys.exit(1)
        finally:
            LOGGER.debug('_event_loop exited')

    async def _read_loop(self):
        """Read messages from the driver and push them onto the event queue.

        This method assumes we're using the stream API. When we use the protocol
        API, this method is not called.
        """
        LOGGER.debug('_read_loop entered')
        while True:
            line = await self.conn.readline()
            if not line:
                break
            # TODO(bfish): If queue is empty, can we just dispatch the event
            # from here?
            self.post_event(load_event(line))
            # Give _event_loop time to run now.
            await asyncio.sleep(0)
        LOGGER.debug('_read_loop exited')

    async def _start(self):
        """Configure the oftr driver based on controller arguments.

        This method runs concurrently with `run()`.
        """
        try:
            await self._get_description()
            if self.args.listen_cert:
                await self._configure_tls()
            if self.args.listen_endpoints:
                await self._listen_on_endpoints()
            # TODO(bfish): Wait for other prestart tasks to finish.
            self._set_phase('START')
        except Exception:  # pylint: disable=broad-except
            LOGGER.exception('Exception in Controller._start')
            self.post_event({'event': 'STARTFAIL'})
            self.post_event({'event': 'EXIT', 'exit_status': 11})

    async def _get_description(self):
        """Check the api version used by oftr.

        Also, check the OpenFlow versions supported.
        """
        try:
            result = await self.rpc_call('OFP.DESCRIPTION')
            # Check API version.
            if float(result['api_version']) != _API_VERSION:
                LOGGER.error('Unsupported API version %s',
                             result['api_version'])
                raise ValueError('Unsupported API version')

            self._supported_versions = result['versions']
            LOGGER.info('Connected to oftr %s', result['sw_desc'])

        except _exc.ControllerException as ex:
            LOGGER.error('Unable to get description from oftr %s', ex)
            raise

    async def _configure_tls(self):
        """Set up a TLS identity for connections to use."""
        try:
            result = await self.rpc_call(
                'OFP.ADD_IDENTITY',
                cert=self.args.listen_cert,
                cacert=self.args.listen_cacert,
                privkey=self.args.listen_privkey)
            # Save tls_id from result so we can pass it in our calls to
            # 'OFP.LISTEN' and 'OFP.CONNECT'.
            self._tls_id = result['tls_id']

        except _exc.ControllerException as ex:
            LOGGER.error('Unable to create TLS identity: %s', ex.message)
            raise

    async def _listen_on_endpoints(self):
        """Listen on a list of endpoints."""
        listen_endpoints = self.args.listen_endpoints
        listen_versions = self.args.listen_versions
        assert isinstance(listen_endpoints, (list, tuple))
        options = ['FEATURES_REQ']
        versions = _prepare_versions(listen_versions, self._supported_versions)
        try:
            for endpt in listen_endpoints:
                result = await self.rpc_call(
                    'OFP.LISTEN',
                    endpoint=endpt,
                    versions=versions,
                    tls_id=self._tls_id,
                    options=options)
                LOGGER.info('Listening on %s [conn_id=%d, versions=%s]', endpt,
                            result['conn_id'], versions)

        except _exc.ControllerException as ex:
            LOGGER.error('Unable to listen on %s: %s', endpt, ex.message)
            raise

    def post_event(self, event):
        """Post an internal event to our event queue."""
        assert isinstance(event, dict)
        self._event_queue.put_nowait(event)

    '''
    def post_message(self, event):
        """Post a message event to our event queue.

        This private api differs from post_event in that we may dispatch the
        event directly when the event queue is empty.
        """
        try:
            if self._event_queue.empty():
                # When there are no other events in line, direct dispatch.
                self._dispatch_event(event)
            else:
                LOGGER.debug('post_message: indirect dispatch: %r', event)
                self._event_queue.put_nowait(event)
        except _exc.StopPropagationException:
            LOGGER.debug('post_message: StopPropagationException caught')
        except Exception:  # pylint: disable=broad-except
            LOGGER.exception('Exception in Controller.post_message')
            sys.exit(1)
    '''

    def _dispatch_event(self, event):
        """Dispatch an event we receive from the queue."""
        LOGGER.debug('_dispatch_event %r', event)
        try:
            if 'event' in event:
                self._handle_internal_event(event)
            elif 'id' in event:
                self._handle_rpc_reply(event)
            elif event['method'] == 'OFP.MESSAGE':
                self._handle_message(event['params'])
            else:
                LOGGER.warning('Unhandled event: %r', event)
        except _exc.StopPropagationException:
            LOGGER.debug('_dispatch_event: StopPropagationException caught')

    def write(self, event, xid=None):
        """Write an event to the output stream.

        If `xid` is specified, return a `_ReplyFuture` to await the response.
        Otherwise, return None.
        """
        self.conn.write(dump_event(event))
        if xid is None:
            return

        # Register future to track the response.
        assert xid > 0
        fut = _ReplyFuture(xid)
        expiration = _timestamp() + _XID_TIMEOUT
        self._reqs[xid] = (fut, expiration, _XID_TIMEOUT)
        return fut

    def rpc_call(self, method, *, ignore_result=False, **params):
        """Send a RPC request and return a future for the reply.

        If ignore_result is True, issue the request but don't return the future.
        """
        if ignore_result:
            xid = None
            event = dict(method=method, params=params)
        else:
            xid = self.next_xid()
            event = dict(id=xid, method=method, params=params)
        if LOGGER.isEnabledFor(logging.DEBUG):
            LOGGER.debug('rpc_call %r', _sanitize_rpc(event))
        return self.write(event, xid)

    def _handle_xid(self, event, xid, except_class=None):
        """Lookup future associated with given xid and give it the event.

        If we find the future, but it has been cancelled, we still return True.

        Return False if no matching xid/future is found.
        """
        if xid not in self._reqs:
            return False

        fut = self._reqs[xid][0]
        if fut.cancelled():
            return True

        if except_class:
            fut.set_exception(except_class(event))
        else:
            fut.set_result(event)

        if not _event_has_more(event) or except_class:
            fut.set_done()
            del self._reqs[xid]

        return True

    def _handle_internal_event(self, event):
        """Called when an internal event is received."""
        event_type = event['event']
        if event_type == 'MESSAGE':
            self._handle_message(event)
            return
        # Immediately begin shutting down if there is an 'EXIT' event.
        if event_type == 'EXIT':
            raise _exc.ExitException(event.get('exit_status', 0))
        # Let apps handle the event.
        for app in self.apps:
            app.handle_event(event, 'event')
        # Check for SIGNAL event asking for exit.
        if event_type == 'SIGNAL' and event['exit']:
            raise _exc.ExitException(_negative_signal_number(event['signal']))

    def _handle_rpc_reply(self, event):
        """Called when a RPC reply is received."""
        result = event.get('result', event)
        except_class = _exc.RPCException if result is event else None
        known_xid = self._handle_xid(result, event['id'], except_class)
        if not known_xid:
            LOGGER.warning('Unrecognized id in RPC reply: %s', event)

    def _handle_message(self, message):
        """Called when a `OFP.MESSAGE` is received."""
        msg_type = message['type']
        if msg_type.startswith('CHANNEL_'):
            self._handle_channel(message)
            return
        if msg_type == 'PACKET_IN' or msg_type == 'PACKET_OUT':
            _convert_pkt(message['msg'])
        except_class = _exc.ErrorException if msg_type == 'ERROR' else None
        # If the message does not have a datapath_id, don't attempt to handle
        # replies based on xid.
        known_xid = False
        if 'datapath_id' in message:
            known_xid = self._handle_xid(message, message['xid'], except_class)

        if not known_xid:
            for app in self.apps:
                app.handle_event(message, 'message')
            # Log all OpenFlow error messages not associated with requests.
            if msg_type == 'ERROR':
                LOGGER.error('ERROR: %r', message)

    def _handle_channel(self, message):
        """Called when `OFP.MESSAGE` is received with type 'CHANNEL_*'."""
        msg_type = message['type']
        if msg_type == 'CHANNEL_ALERT':
            self._handle_alert(message)
            return
        if msg_type == 'CHANNEL_DOWN':
            scope_key = _make_scope_key(message['conn_id'])
            self._cancel_tasks(scope_key)
        for app in self.apps:
            app.handle_event(message, 'message')

    def _handle_alert(self, message):
        """Called when `OFP.MESSAGE` is received with type 'CHANNEL_ALERT'."""
        # First check if this alert was sent in response to something we said.
        msg_xid = message['xid']
        if msg_xid and self._handle_xid(message, msg_xid,
                                        _exc.DeliveryException):
            return
        # Otherwise, we need to report it.
        data_hex = message['data']
        data_len = len(data_hex) / 2
        if len(data_hex) > 100:
            data_hex = '%s...' % data_hex[:100]
        LOGGER.warning(
            'Alert: %s data=%s (%d bytes) [conn_id=%s, datapath_id=%s, xid=%d]',
            message['alert'], data_hex, data_len, message['conn_id'],
            message.get('datapath_id'), msg_xid)

        for app in self.apps:
            app.handle_event(message, 'message')

    async def _idle_task(self):
        """Task to check for requests that have timed out."""
        while True:
            await asyncio.sleep(_IDLE_INTERVAL)
            now = _timestamp()
            timed_out = [
                (xid, fut, timeout)
                for (xid, (fut, expiration, timeout)) in self._reqs.items()
                if expiration <= now
            ]
            for xid, fut, timeout in timed_out:
                if not fut.cancelled():
                    fut.set_exception(_exc.TimeoutException(xid, timeout))
                del self._reqs[xid]

    def _set_phase(self, phase):
        """Called as the run loop changes phase:

            INIT -> PRESTART -> START -> STOP -> POSTFLIGHT.

        N.B. The PREFLIGHT event is delivered at the end of the INIT phase.
        """
        LOGGER.debug('Change phase from "%s" to "%s"', self.phase, phase)
        if self.phase != 'PRESTART':
            self._cancel_tasks(self.phase)
        self.phase = phase
        event = {'event': phase}
        if phase in ('STOP', 'POSTFLIGHT'):
            self._dispatch_event(event)
        elif self._event_queue:
            self.post_event(event)

    def _preflight(self):
        """Called at the end of the INIT phase.
        """
        event = {'event': 'PREFLIGHT'}
        for app in list(self.apps):
            try:
                app.handle_event(event, 'event')
            except _exc.PreflightUnloadException:
                self.apps.remove(app)
            except Exception:  # pylint: disable=broad-except
                app.handle_exception(event, 'event')

    def next_xid(self):
        """Return next xid to use.

        The controller reserves xid 0 and low numbered xid's.
        """

        if self._xid == _MAX_XID:
            self._xid = _MIN_XID
            return self._xid

        self._xid += 1
        return self._xid

    def ensure_future(self, coroutine, *, app=None, datapath_id, conn_id):
        """Run an async coroutine, within the scope of a specific scope_key.

        This function automatically captures exceptions from the coroutine. It
        also cleans up after the task when it is done.
        """

        @functools.wraps(coroutine)
        async def _capture_exception(coroutine):
            try:
                await coroutine
            except asyncio.CancelledError:
                LOGGER.debug('ensure_future cancelled: %r', coroutine)
            except Exception:  # pylint: disable=broad-except
                if app:
                    app.handle_exception(None, scope_key)
                else:
                    LOGGER.error('Exception caught', exc_info=True)

        assert inspect.isawaitable(coroutine)
        task_locals = dict(datapath_id=datapath_id, conn_id=conn_id)
        task = asyncio.ensure_future(_capture_exception(coroutine))
        conn_id = task_locals.get('conn_id')
        if conn_id:
            scope_key = _make_scope_key(conn_id)
        else:
            scope_key = self.phase
        self._tasks[scope_key].append(task)
        task.zof_task_app = app
        task.zof_task_scope = scope_key
        task.zof_task_locals = task_locals
        task.add_done_callback(
            functools.partial(
                self._task_callback, scope_key=scope_key))
        return task

    def _cancel_tasks(self, scope_key):
        """Cancel async tasks associated with the given scope.

        The scope may be a datapath_id or a phase. The done callback removes the
        task from self._tasks.

        If scope_key is 'START' or 'STOP', cancel all tasks.
        """
        LOGGER.debug('_cancel_tasks: scope_key=%s, tasks=%r', scope_key,
                     self._tasks)
        if scope_key == 'START' or scope_key == 'STOP':
            for task_list in self._tasks.values():
                for task in task_list:
                    task.cancel()
        if scope_key in self._tasks:
            for task in self._tasks[scope_key]:
                LOGGER.debug('_cancel_task: %r', task)
                task.cancel()

    def _task_callback(self, task, scope_key):
        """Called when a scoped task is done.
        """
        # LOGGER.debug('_task_callback: %s[%s]', task, scope_key)
        tasks = self._tasks[scope_key]
        tasks.remove(task)
        if not tasks:
            del self._tasks[scope_key]


def _timestamp():
    """Return a monotonic timestamp in seconds."""
    return asyncio.get_event_loop().time()


def _event_has_more(event):
    try:
        return event['type'].startswith('REPLY.') and 'MORE' in event['flags']
    except KeyError:
        return False


class _ReplyFuture:
    """
    Represents the Future-like object returned from OFP.request.

    For simple requests, use it with `await`:

      reply = await OFP.request(SOME_REQUEST, datapath_id=dpid)

    When expecting multiple replies, use it with `async for`:

      async for reply in OFP.request(SOME_REQUEST, datapath_id=dpid):
          process(reply)

    Alternatively, you can call `await` multiple times:

      multi_request = OFP.request(SOME_REQUEST, datapath_id=dpid)
      while not multi_request.done():
          reply = await multi_request
          process(reply)
    """

    def __init__(self, xid):
        self._xid = xid
        self._results = []
        self._future = None
        self._done = False

    def __del__(self):
        if self._results:
            LOGGER.warning('Multiple unread replies xid=%d: %s', self._xid,
                           self._results)

    def cancelled(self):
        return self._future and self._future.cancelled()

    def done(self):
        return not self._results and self._done

    def set_done(self):
        assert not self._done
        self._done = True

    def set_result(self, result):
        assert not isinstance(result, Exception)
        if self._future is None:
            self._results.append(result)
        else:
            self._future.set_result(result)
            self._future = None

    def set_exception(self, exc):
        assert isinstance(exc, Exception)
        if self._future is None:
            self._results.append(exc)
        else:
            self._future.set_exception(exc)
            self._future = None

    def __await__(self):
        """
        If there are existing results, return them immediately. Otherwise, if
        we are still expecting more results, return our future.
        """
        assert self._future is None
        if self._results:
            return _immediate_result(self._results.pop(0)).__await__()
        elif self._done:
            raise asyncio.InvalidStateError(
                'Called "await" too many times on _ReplyFuture')
        else:
            self._future = asyncio.Future()
            return self._future.__await__()

    def __aiter__(self):
        return self

    async def __anext__(self):
        if self.done():
            raise StopAsyncIteration
        return await self


async def _immediate_result(result):
    if isinstance(result, Exception):
        raise result
    return result


def _make_scope_key(conn_id):
    return 'conn_id=%d' % conn_id


def _prepare_versions(listen_versions, supported_versions):
    """Return listen versions."""
    assert supported_versions
    if not listen_versions:
        return supported_versions
    # Check if any desired versions are unsupported.
    unsupported = set(listen_versions) - set(supported_versions)
    if unsupported:
        raise ValueError("Unsupported OpenFlow versions: %r" % unsupported)
    return listen_versions


def _sanitize_rpc(event):
    """Sanitize certain RPC events before logging them."""
    if event['method'] == 'OFP.ADD_IDENTITY':
        event = event.copy()
        event['params'] = event['params'].copy()
        event['params']['privkey'] = '*** ELIDED ***'
    return event


def _negative_signal_number(signame):
    """Return negative number to represent signal named `signame`.

    If signame is unknown, return -99.
    """
    try:
        import signal
        return -getattr(signal, signame)
    except AttributeError:
        return -99


def _convert_pkt(msg):
    """Convert (data, _pkt) to (payload, pkt).

    We also convert msg['data'] from hex string to bytes.

    This method is needed because there's a mismatch between the oftr schema
    and the desired zof schema.
    """
    try:
        # If there's no `data` key, the rest of this is skipped.
        data = bytes.fromhex(msg['data'])
        msg['data'] = data
        # If there's no `_pkt` key, the rest is skipped.
        pkt = pktview_from_list(msg.pop('_pkt'))
        msg['pkt'] = pkt
        # If there's no 'x_pkt_pos' key in pkt, the rest is skipped.
        pkt.payload = data[pkt['x_pkt_pos']:]
    except KeyError:
        pass



import os


class PidFile:
    """Concrete class that represents a PID file."""

    def __init__(self, pid_path):
        self.pid_path = pid_path
        self.exists = False

    def read(self):
        """Read PID file."""
        with open(self.pid_path) as pid_file:
            return int(pid_file.read())

    def write(self):
        """Write PID file."""
        if self.exists or not self.pid_path:
            return
        with open(self.pid_path, 'w') as pid_file:
            pid_file.write(str(os.getpid()))
            self.exists = True

    def remove(self):
        """Remove PID file."""
        if not self.exists:
            return
        try:
            os.unlink(self.pid_path)
        finally:
            self.exists = False

    def __enter__(self):
        self.write()

    def __exit__(self, *args):
        self.remove()



import zof

APP = zof.Application('app_name_here')


@APP.message('packet_in')
def packet_in(event):
    APP.logger.info('packet_in message %r', event)


@APP.message(any)
def other(event):
    APP.logger.info('other message %r', event)


if __name__ == '__main__':
    zof.run()



import re

_ENDPOINT_REGEX = re.compile(r'^(?:(?:\[(\S*)\]|(\S*)):)?(\d+)$')


class Endpoint:
    def __init__(self, endpt):
        self.host, self.port = self._split(endpt)

    @property
    def pair(self):
        """Return 2-tuple containing host and port."""
        return (self.host, self.port)

    @staticmethod
    def _split(endpt):
        m = _ENDPOINT_REGEX.match(endpt)
        if not m:
            raise ValueError('Invalid endpoint: %s' % endpt)
        return (m.group(1) or m.group(2) or '', int(m.group(3)))

    def __str__(self):
        if ':' in self.host:
            return '[%s]:%u' % self.pair
        return '%s:%d' % self.pair



# zof uses the following Exception hierarchy:
#
#     Exception (built into Python)
#       +-- zof.ControllerException
#       |     +-- zof.TimeoutException
#       |     +-- zof.RPCException
#       |     +-- zof.ErrorException
#       |     +-- zof.DeliveryException
#       +-- zof.ControlFlowException
#             +-- zof.StopPropagationException
#             +-- zof.ExitException
#             +-- zof.PreflightUnloadException


class ControllerException(Exception):
    """Base class for exceptions thrown by Controller class."""

    def __init__(self, xid):
        super().__init__()
        self.xid = xid
        self.message = ''


class TimeoutException(ControllerException):
    """Exception that indicates an RPC or OpenFlow request has timed out."""

    def __init__(self, xid, timeout):
        super().__init__(xid)
        self.timeout = timeout

    def __str__(self):
        return '[TimeoutException xid=%s timeout=%s]' % (self.xid,
                                                         self.timeout)


class RPCException(ControllerException):
    """Exception that indicates an error response to an RPC request."""

    def __init__(self, event):
        super().__init__(event['id'])
        self.message = event['error']['message']
        self.code = event['error']['code']

    def __str__(self):
        return '[RPCException xid=%s message=%s]' % (self.xid, self.message)


class ErrorException(ControllerException):
    """Exception that indicates an OpenFlow error reply."""

    def __init__(self, event):
        super().__init__(event['xid'])
        self.event = event

    def __str__(self):
        return '[ErrorException xid=%s event=%s]' % (self.xid, self.event)


class DeliveryException(ControllerException):
    """Exception that indicates an OpenFlow message couldn't be delivered."""

    def __init__(self, event):
        super().__init__(event['xid'])
        self.event = event

    def __str__(self):
        return '[DeliveryException xid=%s event=%s]' % (self.xid, self.event)


class ControlFlowException(Exception):
    """Base class for control flow exceptions used in zof."""


class StopPropagationException(ControlFlowException):
    """Exception class used by an app to prevent handlers in other apps from
    executing.
    """

    def __str__(self):
        return '[StopPropagationException]'


class ExitException(ControlFlowException):
    """Internal exception class used to stop the controller."""

    def __init__(self, exit_status=0):
        super().__init__()
        self.exit_status = exit_status

    def __str__(self):
        return '[ExitException exit_status=%d]' % self.exit_status


class PreflightUnloadException(ControlFlowException):
    """Exception class used by an app unload itself. This exception should
    only be raised in a 'preflight' event handler.
    """

    def __str__(self):
        return '[PreflightUnloadException]'


class CodecError(ValueError):
    """Exception class used in codec failures.


    """

    def __init__(self, msg):
        super().__init__(msg, None)

    def __str__(self):
        return self.args[0]



import asyncio
import time
import unittest
from zof.connection import Connection
from .asynctestcase import AsyncTestCase

# Max message size is 1MB.
MSG_LIMIT = 2**20

# Skip tests that send big messages.
SKIP_BIG_MESSAGES = False

# Pass these args when launching oftr.
OFTR_ARGS = ''  # '--trace=rpc,msg --loglevel=debug'


class ConnectionTestCase(AsyncTestCase):
    async def setUp(self):
        oftr_options = {'args': OFTR_ARGS}
        self.conn = Connection(oftr_options=oftr_options)
        await self.conn.connect()

    async def tearDown(self):
        self.conn.close()
        return_code = await self.conn.disconnect()
        if return_code:
            raise Exception('oftr exited with return code %d' % return_code)

    # Basic tests.

    async def test_rpc(self):
        # If we write a valid JSON-RPC method, we should get a result back.
        msg = b'{"id":1234,"method":"OFP.DESCRIPTION"}'
        self.conn.write(msg)
        result = await self.conn.readline()
        self.assert_prefix(
            result, b'{"id":1234,"result":{"api_version":"0.9","sw_desc":')

    async def test_rpc_no_id(self):
        # If we write a valid JSON-RPC method with no id, there should be no response.
        msg = b'{"method":"OFP.DESCRIPTION"}'
        self.conn.write(msg)
        with self.assertRaises(asyncio.TimeoutError):
            result = await asyncio.wait_for(self.conn.readline(), 1.0)
            self.assertIsNone(result)

    async def test_rpc_send_no_datapath(self):
        # If we write a valid JSON-RPC OFP.SEND method, there should be a response.
        msg = b'{"id":7,"method":"OFP.SEND","params":{"version":4,"type":"FEATURES_REQUEST"}}'
        self.conn.write(msg)
        result = await self.conn.readline()
        self.assert_prefix(
            result,
            b'{"id":7,"error":{"code":-65000,"message":"YAML:1:39: error: unable to locate connection; no conn_id or datapath_id'
        )

    async def test_rpc_send_no_id_datapath(self):
        msg = b'{"method":"OFP.SEND","params":{"version":4,"type":"FEATURES_REQUEST","xid":12}}'
        self.conn.write(msg)
        result = await self.conn.readline()
        self.assertRegex(
            result,
            b'{"method":"OFP.MESSAGE","params":{"type":"CHANNEL_ALERT","time":"[0-9.]+","conn_id":0,"xid":12,'
        )

    async def test_rpc_send_missing_datapath(self):
        # If we write a valid JSON-RPC OFP.SEND method, there should be a response.
        msg = b'{"id":8,"method":"OFP.SEND","params":{"datapath_id":"00:00:00:00:00:00:00:01","type":"FEATURES_REQUEST"}}'
        self.conn.write(msg)
        result = await self.conn.readline()
        self.assert_prefix(
            result,
            b'{"id":8,"error":{"code":-65000,"message":"YAML:1:39: error: unable to locate datapath_id 00:00:00:00:00:00:00:01'
        )

    async def test_rpc_hex_id(self):
        # If we write a valid JSON-RPC method with a hexadecimal ID, we should
        # get a result back.
        msg = b'{"id":"0x1234","method":"OFP.DESCRIPTION"}'
        self.conn.write(msg)
        result = await self.conn.readline()
        self.assert_prefix(
            result, b'{"id":4660,"result":{"api_version":"0.9","sw_desc":')

    # Test the connection transport limits.

    async def test_short_malformed(self):
        # Write a short malformed json message.
        self.conn.write(b'hello')
        result = await self.conn.readline()
        self.assertEqual(
            result,
            b'{"id":null,"error":{"code":-32600,"message":"YAML:1:1: error: not a mapping\\nhello\\n^~~~~"}}'
        )
        await self._still_alive()

    @unittest.skipIf(SKIP_BIG_MESSAGES, "skipping big message tests")
    async def test_write_max_size(self):
        # Write a maximum sized message.
        junk = b'x' * (MSG_LIMIT - 1)
        self.conn.write(junk)
        result = await self.conn.readline()
        self.assert_prefix(
            result,
            b'{"id":null,"error":{"code":-32600,"message":"YAML:1:1: error: not a mapping\\nxxx'
        )
        await self._still_alive()

    @unittest.skipIf(SKIP_BIG_MESSAGES, "skipping big message tests")
    async def test_write_too_big(self):
        # Write a message that is too big.
        junk = b'x' * MSG_LIMIT
        self.conn.write(junk)

        # We should receive a message that the RPC request is too big.
        result = await self.conn.readline()
        self.assertEqual(
            result,
            b'{"id":null,"error":{"code":-32600,"message":"RPC request is too big"}}'
        )
        await self._end_of_input()

    @unittest.skipIf(SKIP_BIG_MESSAGES, "skipping big message tests")
    async def test_zmassive_write(self):
        # Now, write enough messages to really back things up.
        count = 5
        junk = b'z' * (MSG_LIMIT - 1)

        for _ in range(count):
            self.conn.write(junk)

        t = time.perf_counter()
        await self.conn.drain()
        t = time.perf_counter() - t
        print('drain time elapsed', t)

        for _ in range(count):
            result = await self.conn.readline()
            self.assert_prefix(
                result,
                b'{"id":null,"error":{"code":-32600,"message":"YAML:1:1: error: not a mapping\\nzzz'
            )

        await self._still_alive()

    # Test invalid input.

    async def test_rpc_invalid_json(self):
        # If we write an invalid message, we should get back an error with a
        # id=null.
        msg = b'{invalid'
        self.conn.write(msg)
        result = await self.conn.readline()
        self.assert_prefix(result,
                           b'{"id":null,"error":{"code":-32600,"message":')
        await self._still_alive()

    async def test_rpc_incorrect_jsonrpc_version(self):
        # If we write a message with an incorrect jsonrpc version, we should
        # get back an error.
        msg = b'{"jsonrpc":"3.0","id":"13","method":"OFP.DESCRIPTION"}'
        self.conn.write(msg)
        result = await self.conn.readline()
        self.assert_prefix(
            result,
            b'{"id":13,"error":{"code":-32600,"message":"YAML:1:2: error: Unsupported jsonrpc version'
        )
        await self._still_alive()

    async def test_rpc_invalid_id(self):
        msg = b'{"id":"abc","method":"OFP.DESCRIPTION"}'
        self.conn.write(msg)
        result = await self.conn.readline()
        self.assert_prefix(
            result,
            b'{"id":null,"error":{"code":-32600,"message":"YAML:1:7: error: Invalid RPC ID'
        )
        await self._still_alive()

    async def test_rpc_id_null(self):
        # If we write a message with a null id, it is treated as a valid id.
        msg = b'{"id":null,"method":"OFP.DESCRIPTION"}'
        self.conn.write(msg)
        result = await self.conn.readline()
        self.assert_prefix(result, b'{"id":null,"result":{"api_version":')
        await self._still_alive()

    async def test_rpc_id_too_big(self):
        for id_val in [2**64 - 2, 2**64 - 1, 2**64, 2**64 + 1]:
            msg = b'{"id":%d,"method":"OFP.DESCRIPTION"}' % id_val
            self.conn.write(msg)
            result = await self.conn.readline()
            self.assert_prefix(
                result,
                b'{"id":null,"error":{"code":-32600,"message":"YAML:1:7: error: Invalid RPC ID'
            )
            await self._still_alive()

    async def test_rpc_id_max(self):
        id_val = 2**64 - 3
        msg = b'{"id":%d,"method":"OFP.DESCRIPTION"}' % id_val
        self.conn.write(msg)
        result = await self.conn.readline()
        self.assert_prefix(result,
                           b'{"id":%d,"result":{"api_version":' % id_val)
        await self._still_alive()

    async def test_ofp_send_invalid(self):
        # If we write an invalid OFP.SEND message, we should get back an error.
        msg = b'{"id":321,"method":"OFP.SEND","params":{"type":"foo"}}'
        self.conn.write(msg)
        result = await asyncio.wait_for(self.conn.readline(), 1.0)
        self.assert_prefix(
            result,
            b'{"id":321,"error":{"code":-32600,"message":"YAML:1:48: error: unknown value \\"foo\\" Did you mean \\"HELLO\\"?'
        )
        await self._still_alive()

    async def test_ofp_send_invalid_no_id(self):
        # If we write an invalid OFP.SEND message with no id, we should get back
        # a OFP.MESSAGE notification.
        msg = b'{"method":"OFP.SEND","params":{"type":"foo","xid":133}}'
        self.conn.write(msg)
        result = await asyncio.wait_for(self.conn.readline(), 1.0)
        self.assertRegex(
            result,
            b'{"method":"OFP.MESSAGE","params":{"type":"CHANNEL_ALERT","time":"[0-9.]+","conn_id":0,"xid":133,'
        )
        await self._still_alive()

    async def test_ofp_send_invalid_no_alert(self):
        # If we write an invalid OFP.SEND message for a non-existant datapath
        # with 'NO_ALERT' specified, we should not get a response back.
        msg = b'{"method":"OFP.SEND","params":{"datapath_id":"0x01","type":"FEATURES_REQUEST","xid":134,"flags":["NO_ALERT"]}}'
        self.conn.write(msg)
        with self.assertRaises(asyncio.TimeoutError):
            result = await asyncio.wait_for(self.conn.readline(), 1.0)
            self.assertIsNone(result)

    # The following tests are taken from http://www.jsonrpc.org/specification

    async def test_rpc_call_of_nonexistant_method(self):
        msg = b'{"jsonrpc": "2.0", "method": "foobar", "id": "1"}'
        self.conn.write(msg)
        result = await self.conn.readline()
        self.assert_prefix(
            result,
            b'{"id":1,"error":{"code":-32601,"message":"YAML:1:30: error: unknown method'
        )
        await self._still_alive()

    async def test_rpc_call_with_invalid_json(self):
        msg = b'{"jsonrpc": "2.0", "method": "foobar, "params": "bar", "baz]'
        self.conn.write(msg)
        result = await self.conn.readline()
        self.assert_prefix(result,
                           b'{"id":null,"error":{"code":-32600,"message":')
        await self._still_alive()

    async def test_rpc_call_with_invalid_request_object(self):
        msg = b'{"jsonrpc": "2.0", "method": 1, "params": "bar"}'
        self.conn.write(msg)
        result = await self.conn.readline()
        self.assert_prefix(result,
                           b'{"id":null,"error":{"code":-32600,"message":')
        await self._still_alive()

    async def test_rpc_call_with_invalid_batch(self):
        msg = b'[1,2,3]'
        self.conn.write(msg)
        result = await self.conn.readline()
        self.assert_prefix(result,
                           b'{"id":null,"error":{"code":-32600,"message":')
        await self._still_alive()

    # Test close(write=True)

    async def test_close_write(self):
        self.conn.write(b'{"id":12300,"method":"OFP.DESCRIPTION"}')
        self.conn.close(write=True)
        result = await self.conn.readline()
        self.assert_prefix(result, b'{"id":12300,"result":{"api_version":')
        await self._end_of_input()

    async def test_close_incomplete_write(self):
        self.conn.write(
            b'{"id":12300,"method":"OFP.DESCRIPTION"}', delimiter=None)
        self.conn.close(write=True)
        result = await self.conn.readline()
        self.assert_prefix(result, b'')
        await self._end_of_input()

    # Helper methods

    async def _end_of_input(self):
        result = await self.conn.readline()
        self.assertEqual(result, b'')

    async def _still_alive(self):
        self.conn.write(b'{"id":12300,"method":"OFP.DESCRIPTION"}')
        result = await self.conn.readline()
        self.assert_prefix(result, b'{"id":12300,"result":{"api_version":')

    def assert_prefix(self, value, prefix):
        if not value.startswith(prefix):
            self.fail('Prefix %s does not match: %s' % (prefix, value))


class ConnectionMiscTestCase(unittest.TestCase):
    def test_find_oftr_path(self):
        path = Connection.find_oftr_path()
        self.assertTrue(path)


class ConnectionProtocolTestCase(AsyncTestCase):
    async def setUp(self):
        oftr_options = {'args': '--trace=rpc'}
        self.queue = asyncio.Queue()
        self.conn = Connection(oftr_options=oftr_options)
        await self.conn.connect(self.queue.put_nowait)

    async def tearDown(self):
        self.conn.close()
        return_code = await self.conn.disconnect()
        if return_code:
            raise Exception('oftr exited with return code %d' % return_code)

    async def test_rpc(self):
        # If we write a valid JSON-RPC method, we should get a result back
        # on the queue.
        msg = b'{"id":1237,"method":"OFP.DESCRIPTION"}'
        self.conn.write(msg)
        result = await self.queue.get()
        self.assertIsInstance(result, dict)
        self.assertEqual(result['id'], 1237)
        self.assertEqual(result['result']['api_version'], '0.9')



import unittest
from zof.event import dump_event


class DumpEventTestCase(unittest.TestCase):
    def test_empty_dict(self):
        # Test empty event as Python object.
        data = dump_event({})
        self.assertIsInstance(data, bytes)
        self.assertEqual(data, b'{}')

    def test_hello(self):
        # Test hello event as Python object.
        data = dump_event(dict(msg=dict(type='foo')))
        self.assertIsInstance(data, bytes)
        self.assertEqual(data, b'{"msg":{"type":"foo"}}')

    def test_string(self):
        # Test event as YAML string.
        data = dump_event('type: string\nmsg: "{}"')
        self.assertIsInstance(data, bytes)
        self.assertEqual(data, b'type: string\nmsg: "{}"')

    def test_bytes(self):
        # Test that bytes values are converted to hex strings.
        data = dump_event({'x': b'12345'})
        self.assertEqual(data, b'{"x":"3132333435"}')

    def test_nonserializable(self):
        # Test a value that isn't JSON serializable.
        with self.assertRaisesRegex(TypeError,
                                    "Value .+ is not JSON serializable"):
            dump_event({'x': set()})

    def test_utf8(self):
        # Test that Unicode/UTF-8 works.
        data = dump_event({'x': '\u20AC\U00010302\U0010fffd'})
        self.assertEqual(
            data, b'{"x":"\xe2\x82\xac\xf0\x90\x8c\x82\xf4\x8f\xbf\xbd"}')

    def test_getstate(self):
        # Test custom class with __getstate__.
        class Foo:
            def __init__(self):
                self.y = 1

            def __getstate__(self):
                return self.__dict__

        data = dump_event({'x': Foo()})
        self.assertEqual(data, b'{"x":{"y":1}}')



"""Implements Connection class."""

import asyncio
import logging
import shutil
import shlex
from zof.protocol import Protocol

_DEFAULT_LIMIT = 2**20  # max line length is 1MB

LOGGER = logging.getLogger(__package__)


class Connection(object):
    """Concrete class representing a connection to the `oftr` driver.

    This class supports both the stream and protocol API's. If you pass a
    'post_message' function argument to `connect()`, you will use the protocol
    API which may be faster.

    Keyword Args:
        oftr_options (Dict[str, str]): Dictionary with options:
                path: Path to oftr executable (default=<local system path>)
                subcmd: Subcommand name (default='jsonrpc')
                args: Command line arguments for oftr (default='')
                prefix: Command line prefix for launching oftr (default='')

    The command to execute oftr are constructed from oftr_options:

        "<prefix> <path> <subcmd> <args>"
    """

    def __init__(self, *, oftr_options=None):
        if oftr_options is None:
            oftr_options = {}
        # Verify that connection options are all strings or None.
        for key, value in oftr_options.items():
            if value is not None and not isinstance(value, str):
                raise ValueError('Unexpected oftr option: %s=%r' %
                                 (key, value))
        self._conn = None
        self._input = None
        self._output = None
        self._protocol = None
        self._pid = None

        oftr_path = self.find_oftr_path(oftr_options.get('path'))
        oftr_subcmd = oftr_options.get('subcmd') or 'jsonrpc'
        oftr_args = oftr_options.get('args') or ''
        oftr_prefix = oftr_options.get('prefix') or ''

        # Construct the cmd used to launch oftr subprocess.
        cmd = '%s %s %s %s' % (oftr_prefix, oftr_path, oftr_subcmd, oftr_args)
        self._oftr_cmd = shlex.split(cmd)
        assert self._oftr_cmd

    @property
    def pid(self):
        """Return process ID for oftr process.

        Returns:
            (int) process id of oftr process (or None if not running)
        """
        return self._pid

    async def connect(self, post_message=None):
        """Set up connection to the oftr driver.

        If the 'post_message' argument is present, use the faster protocol api.

        Args:
            post_message (function): single arg function to post received
                message events
        Returns:
            (int) process id of oftr process
        """
        # If a callback is provided, use the asyncio protocol api.
        if post_message:
            return await self._connect_protocol(post_message)

        LOGGER.debug("Launch oftr %r (stream API)", self._oftr_cmd)

        try:
            # When we create the subprocess, make it a session leader.
            # We do not want SIGINT signals sent from the terminal to reach
            # the subprocess.
            proc = await asyncio.create_subprocess_exec(
                *self._oftr_cmd,
                stdin=asyncio.subprocess.PIPE,
                stdout=asyncio.subprocess.PIPE,
                limit=_DEFAULT_LIMIT,
                start_new_session=True)
            self._conn = proc
            self._input = proc.stdout
            self._output = proc.stdin
            self._pid = proc.pid
            return self._pid

        except (PermissionError, FileNotFoundError):
            LOGGER.error('Unable to find executable: "%r"', self._oftr_cmd)
            raise

    async def _connect_protocol(self, post_message):
        """Set up connection to oftr driver (using the Protocol api).

        Returns:
            (int) process id of oftr process
        """
        LOGGER.debug("Launch oftr %r (protocol API)", self._oftr_cmd)

        try:
            # When we create the subprocess, make it a session leader.
            # We do not want SIGINT signals sent from the terminal to reach
            # the subprocess.
            loop = asyncio.get_event_loop()
            transport, protocol = await loop.subprocess_exec(
                lambda: Protocol(post_message),
                *self._oftr_cmd,
                stderr=None,
                start_new_session=True)
            self._conn = transport
            self._protocol = protocol
            self._input = transport.get_pipe_transport(1)
            self._output = transport.get_pipe_transport(0)
            self._pid = transport.get_pid()
            return self._pid

        except (PermissionError, FileNotFoundError):
            LOGGER.error('Unable to find executable: "%r"', self._oftr_cmd)
            raise

    async def disconnect(self):
        """Wait for oftr connection to close.
        """
        if self._conn is None:
            return 0
        if self._protocol:
            await self._protocol.exit_future
            self._conn.close()
            return_code = self._conn.get_returncode()
        else:
            return_code = await self._conn.wait()
        if return_code:
            LOGGER.error("oftr exited with return code %s", return_code)
        self._input = None
        self._output = None
        self._conn = None
        self._protocol = None
        self._pid = None
        return return_code

    async def readline(self, delimiter=b'\x00'):
        """Read next incoming line from the connection.
        """
        try:
            result = await self._input.readuntil(delimiter)
            return result[0:-1]
        except asyncio.streams.IncompleteReadError as ex:
            if ex.partial:
                LOGGER.warning('oftr incomplete read: %d bytes ignored',
                               len(ex.partial))
            return b''

    def write(self, data, delimiter=b'\x00'):
        """Write data to the connection.
        """
        self._output.write(data + delimiter if delimiter else data)

    async def drain(self):
        """Wait while the output buffer is flushed.
        """
        LOGGER.info('oftr connection drain: buffer_size=%d',
                    self.get_write_buffer_size())
        return await self._output.drain()

    def close(self, write=False):
        """Close the connection.
        """
        if self._conn is None:
            return
        if write:
            self._output.close()
        else:
            try:
                self._conn.terminate()
            except ProcessLookupError:
                # Ignore failure when process already died.
                pass

    def get_write_buffer_size(self):
        """Get size of the write buffer.
        """
        return self._output.transport.get_write_buffer_size()

    @staticmethod
    def find_oftr_path(default=None):
        """Return path to oftr executable.
        """
        if default:
            return default
        path = shutil.which('oftr')
        if not path:
            raise RuntimeError('Unable to find oftr executable.')
        return path



import zof

APP = zof.Application('hub')


@APP.message('channel_up')
def channel_up(event):
    APP.logger.info('Channel UP datapath_id=%s', event['datapath_id'])
    FLOW_MOD_TABLE_MISS.send()


@APP.message('packet_in')
def packet_in(event):
    APP.logger.info('Packet_in %r', event)
    msg = event['msg']
    PACKET_OUT.send(in_port=msg['in_port'], data=msg['data'])


FLOW_MOD_TABLE_MISS = zof.compile('''
  type: FLOW_MOD
  msg:
    table_id: 0
    command: ADD
    priority: 0
    match:
    instructions:
      - instruction: APPLY_ACTIONS
        actions:
          - action: OUTPUT
            port_no: CONTROLLER
            max_len: NO_BUFFER
''')

PACKET_OUT = zof.compile('''
  type: PACKET_OUT
  msg:
    in_port: $in_port
    actions:
      - action: OUTPUT
        port_no: ALL
    data: $data
''')

if __name__ == '__main__':
    zof.run()



import unittest
from zof.datapath import DatapathList, Datapath, normalize_datapath_id, normalize_port_no
from zof.objectview import to_json


class DatapathTestCase(unittest.TestCase):
    def test_datapath_list(self):
        dpids = DatapathList()
        self.assertFalse(dpids)
        self.assertEqual(len(dpids), 0)

        dp1 = dpids.add_datapath(
            datapath_id='00:00:00:00:00:00:00:01', conn_id=1)
        dp2 = dpids.add_datapath(
            datapath_id='00:00:00:00:00:00:00:02', conn_id=2)
        self.assertTrue(dpids)
        self.assertEqual(len(dpids), 2)
        self.assertIs(dpids[0x01], dp1)
        self.assertIs(dpids['2'], dp2)

        dpids.delete_datapath(datapath_id=dp1.datapath_id)
        self.assertEqual(len(dpids), 1)

        dps = [dp for dp in dpids]
        self.assertEqual(dps, [dp2])

    def test_datapath(self):
        dp1 = Datapath(datapath_id='00:00:00:00:00:00:00:01', conn_id=1001)
        self.assertEqual(dp1.datapath_id, '00:00:00:00:00:00:00:01')
        self.assertEqual(dp1.conn_id, 1001)
        self.assertTrue(dp1)  # Empty datapath is still "true"

        port1 = dp1.add_port(port_no=1)
        port2 = dp1.add_port(port_no=2)
        port3 = dp1.add_port(port_no='LOCAL')
        self.assertEqual(len(dp1), 3)

        self.assertIs(dp1['2'], port2)
        self.assertIs(dp1['local'], port3)

        ports = [port for port in dp1]
        self.assertEqual(ports, [port1, port2, port3])

        self.assertEqual(port1.datapath_id, '00:00:00:00:00:00:00:01')
        self.assertIs(port1.datapath, dp1)
        self.assertEqual(port1.port_no, 1)

    def test_port(self):
        dp1 = Datapath(datapath_id='00:00:00:00:00:00:00:01', conn_id=1001)
        port1 = dp1.add_port(port_no=1)

        self.assertEqual(port1.port_no, 1)
        self.assertIs(port1.datapath, dp1)
        self.assertTrue(port1.up)
        self.assertFalse(port1.admin_down)

    def test_add_datapath(self):
        dpids = DatapathList()

        # add_datapath is idempotent, if values are identical.
        dp1 = dpids.add_datapath(datapath_id=0x01, conn_id=0x02)
        dp2 = dpids.add_datapath(datapath_id=0x01, conn_id=0x02)
        self.assertEqual(len(dpids), 1)
        self.assertIs(dp2, dp1)

        with self.assertRaises(ValueError):
            # add_datapath fails if conn_id values differ
            dpids.add_datapath(datapath_id=0x01, conn_id=0x03)

        # We don't check that conn_id is unique.
        dp3 = dpids.add_datapath(datapath_id=0x02, conn_id=0x02)
        self.assertIsNot(dp3, dp1)

    def test_delete_datapath(self):
        dpids = DatapathList()
        dp1 = dpids.add_datapath(datapath_id=0x01, conn_id=0x01)
        dp2 = dpids.add_datapath(datapath_id=0x02, conn_id=0x02)

        result = dpids.delete_datapath(datapath_id=1)
        self.assertIs(result, dp1)
        self.assertEqual(len(dpids), 1)
        result = dpids.delete_datapath(datapath_id=1)
        self.assertIsNone(result)

        result = dpids.delete_datapath(datapath_id=dp2.datapath_id)
        self.assertIs(result, dp2)
        self.assertEqual(len(dpids), 0)

        result = dpids.delete_datapath(datapath_id=12345)
        self.assertIsNone(result)

    def test_repr(self):
        dpids = DatapathList()
        dp1 = dpids.add_datapath(datapath_id=0x01, conn_id=0x01)
        self.assertEqual(repr(dp1), "'<zof.Datapath 1>'")
        self.assertEqual(repr(dp1), "'%s'" % dp1.__getstate__())

        port1 = dp1.add_port(port_no=2)
        self.assertEqual(repr(port1), "'<zof.Port 2>'")
        self.assertEqual(repr(port1), "'%s'" % port1.__getstate__())

    def test_add_port(self):
        dp1 = Datapath(datapath_id='00:00:00:00:00:00:00:01', conn_id=1001)

        # add_port is idempotent if port_no's are identical.
        port1 = dp1.add_port(port_no=1)
        port2 = dp1.add_port(port_no=1)
        self.assertEqual(len(dp1), 1)
        self.assertIs(port1, port2)

    def test_delete_port(self):
        dp1 = Datapath(datapath_id='00:00:00:00:00:00:00:01', conn_id=1001)
        port1 = dp1.add_port(port_no=1)
        port2 = dp1.add_port(port_no=2)

        result = dp1.delete_port(port_no=1)
        self.assertIs(result, port1)
        self.assertEqual(len(dp1), 1)
        result = dp1.delete_port(port_no=1)
        self.assertIsNone(result)

        result = dp1.delete_port(port_no=2)
        self.assertIs(result, port2)
        self.assertEqual(len(dp1), 0)

        result = dp1.delete_port(port_no=12345)
        self.assertIsNone(result)

    def test_add_ports(self):
        dp1 = Datapath(datapath_id='00:00:00:00:00:00:00:01', conn_id=1001)
        port1_desc = {
            'port_no': 9,
            'hw_addr': '00:00:00:00:00:01',
            'name': 'Port 1',
            'state': ['LINK_DOWN'],
            'config': ['PORT_DOWN']
        }
        dp1.add_ports([port1_desc])

        port1 = dp1[9]
        self.assertEqual(port1.port_no, 9)
        self.assertEqual(port1.hw_addr, '00:00:00:00:00:01')
        self.assertEqual(port1.name, 'Port 1')
        self.assertFalse(port1.up)
        self.assertTrue(port1.admin_down)

    def test_normalize_datapath(self):
        dpid = normalize_datapath_id('00:00:00:00:00:00:00:01')
        self.assertEqual(dpid, 1)

        dpid = normalize_datapath_id('ff:ff:ff:ff:ff:ff:ff:ff')
        self.assertEqual(dpid, 2**64 - 1)

        dpid = normalize_datapath_id(0x12345)
        self.assertEqual(dpid, 0x12345)

        dpid = normalize_datapath_id('0x567')
        self.assertEqual(dpid, 0x567)

        with self.assertRaises(ValueError):
            normalize_datapath_id('blah')

        with self.assertRaises(ValueError):
            normalize_datapath_id('')

    def test_normalize_port_no(self):
        port_no = normalize_port_no(123)
        self.assertEqual(port_no, 123)

        port_no = normalize_port_no('0x123')
        self.assertEqual(port_no, 0x123)

        port_no = normalize_port_no('LOCAL')
        self.assertEqual(port_no, 'LOCAL')

        port_no = normalize_port_no('controller')
        self.assertEqual(port_no, 'CONTROLLER')

    def test_json(self):
        dp1 = Datapath(datapath_id='00:00:00:00:00:00:00:01', conn_id=1001)
        result = to_json({'datapath': dp1})
        self.assertEqual(
            result, '{"datapath":"<zof.Datapath 00:00:00:00:00:00:00:01>"}')

        result = repr({'datapath': dp1})
        self.assertEqual(
            result, "{'datapath': '<zof.Datapath 00:00:00:00:00:00:00:01>'}")



import zof

APP = zof.Application('table_features', exception_fatal=True)

DESC = zof.compile('''
type: REQUEST.DESC
''')

TABLE_FEATURES = zof.compile('''
type: REQUEST.TABLE_FEATURES
msg: []
''')


async def table_features_match_fields1():
    """Get set of all fields available in OF matches.
    """
    fields = set()
    tfm_count = 0
    entry_count = 0

    async for tfm in TABLE_FEATURES.request():
        tfm_count += 1
        for tfm_entry in tfm['msg']:
            entry_count += 1
            for field in tfm_entry['match']:
                fields.add(field)

    print('Processed %d entries in %d TableFeatures messages' %
          (entry_count, tfm_count))
    return fields


async def table_features_match_fields2():
    """Get set of all fields available in OF matches.
    """
    fields = set()
    tfm_count = 0
    entry_count = 0

    tfm = await TABLE_FEATURES.request()
    tfm_count += 1
    for tfm_entry in tfm['msg']:
        entry_count += 1
        for field in tfm_entry['match']:
            fields.add(field)

    print('Processed %d entries in %d TableFeatures messages' %
          (entry_count, tfm_count))
    return fields


@APP.message('channel_up')
async def channel_up(event):
    info = await DESC.request()
    print('Description: %s %s' %
          (info['msg']['hw_desc'], info['msg']['sw_desc']))

    fields = await table_features_match_fields1()
    for field in sorted(list(fields)):
        print(field)

    fields2 = await table_features_match_fields2()
    assert fields2 == fields

    zof.post_event({'event': 'EXIT'})


if __name__ == '__main__':
    zof.run()



from .objectview import to_json, from_json


def load_event(event):
    try:
        return from_json(event)
    except ValueError as ex:
        # Report malformed JSON input.
        return {'event': 'EXCEPTION', 'reason': str(ex), 'input': event}


def dump_event(event):
    if isinstance(event, str):
        return event.encode('utf-8')
    return to_json(event).encode('utf-8')



import unittest
from ipaddress import IPv4Address, IPv6Address, ip_address
from zof.pktview import make_pktview, pktview_from_list, pktview_to_list, PktView, pktview_alias, pktview_from_ofctl, convert_slash_notation


class PktViewTestCase(unittest.TestCase):
    def test_init(self):
        pkt = make_pktview(a=1, b=2)
        self.assertEqual(pkt.a, 1)
        self.assertEqual(pkt.b, 2)

    def test_basic(self):
        data = [dict(field='A', value='a'), dict(field='B', value=2)]
        pkt = pktview_from_list(data)
        self.assertEqual(pkt.a, 'a')
        self.assertEqual(pkt.b, 2)

        fields = pktview_to_list(pkt)
        fields.sort(key=_by_field)
        self.assertEqual(fields, data)

    def test_invalid_init(self):
        with self.assertRaises(ValueError):
            pktview_from_list(dict(A='a'))

    def test_from_list(self):
        data = [dict(field='A', value=5, mask=255)]
        pkt = pktview_from_list(data)
        self.assertEqual(pkt, {'a': (5, 255)})

        pkt = pktview_from_list(data, slash_notation=True)
        self.assertEqual(pkt, {'a': '5/255'})

        with self.assertRaises(ValueError):
            pktview_from_list([{'field': 'PAYLOAD', 'value': '1234'}])

    def test_to_list(self):
        data = dict(a='a', b=2, c='ddd')
        fields = pktview_to_list(data)
        fields.sort(key=_by_field)
        self.assertEqual(
            fields, [
                dict(
                    field='A', value='a'), dict(
                        field='B', value=2), dict(
                            field='C', value='ddd')
            ])

        with self.assertRaisesRegex(ValueError, r'len\(tuple\) != 2'):
            pktview_to_list({'a': (1, 2, 3)})

    def test_alias_attr(self):
        pkt = make_pktview()
        pkt.hop_limit = 5
        self.assertEqual(pkt.nx_ip_ttl, 5)
        self.assertEqual(pkt['nx_ip_ttl'], 5)
        self.assertEqual(pkt['hop_limit'], 5)

        pkt.nx_ip_ttl = 10
        self.assertEqual(pkt.hop_limit, 10)
        self.assertTrue('hop_limit' in pkt)
        del pkt.hop_limit
        self.assertFalse('nx_ip_ttl' in pkt)

    def test_alias_subscript(self):
        pkt = make_pktview()
        pkt['hop_limit'] = 6
        self.assertEqual(pkt['hop_limit'], 6)
        self.assertEqual(pkt.hop_limit, 6)
        self.assertEqual(pkt.nx_ip_ttl, 6)
        self.assertTrue('hop_limit' in pkt)
        self.assertTrue('nx_ip_ttl' in pkt)
        del pkt['hop_limit']
        self.assertFalse('hop_limit' in pkt)
        self.assertFalse('nx_ip_ttl' in pkt)
        with self.assertRaisesRegex(KeyError, 'hop_limit'):
            del pkt['hop_limit']

    def test_get_protocol(self):
        pkt = make_pktview(ipv4_src='1.2.3.4', eth_type=0x0800)
        self.assertEqual(pkt.get_protocol('IPV4'), pkt)
        self.assertFalse(pkt.get_protocol('IPV6'))
        self.assertFalse(pkt.get_protocol('ARP'))
        self.assertEqual(pkt.get_protocol('ethernet'), pkt)

    def test_pktview_alias_identity(self):
        # Test that identity alias doesn't recurse forever.
        class SubPktView(PktView):
            ip = pktview_alias('ip')

        pkt = SubPktView({})
        pkt.ip = '1.2.3.4'
        self.assertEqual(pkt.ip, '1.2.3.4')
        self.assertTrue('ip' in pkt)
        del pkt.ip
        self.assertFalse('ip' in pkt)

    def test_pktview_alias_converter(self):
        class SubPktView(PktView):
            ip = pktview_alias('ip', ip_address)

        pkt = SubPktView({})
        pkt.ip = '1.2.3.4'
        self.assertIsInstance(pkt.ip, IPv4Address)
        self.assertEqual(pkt.ip, ip_address('1.2.3.4'))
        self.assertTrue('ip' in pkt)
        del pkt.ip
        self.assertFalse('ip' in pkt)

    def test_pktview_src(self):
        pkt = make_pktview(ipv4_src='1.2.3.4')
        self.assertEqual(pkt.src, '1.2.3.4')
        with self.assertRaises(AttributeError):
            self.assertFalse(pkt.dst)

        pkt = make_pktview(eth_src='00:00:00:00:00:01')
        with self.assertRaises(AttributeError):
            self.assertFalse(pkt.src)

    def test_pktview_get(self):
        pkt = make_pktview(ipv4_src='1.2.3.4', eth_type=0x0800)
        self.assertEqual(pkt.get('ipv4_src'), '1.2.3.4')
        self.assertEqual(pkt.get('ipv4_dst', '5.6.7.8'), '5.6.7.8')

    def test_pktview_items(self):
        pkt = make_pktview(ipv4_src='1.2.3.4', eth_type=0x0800)
        items = list(pkt.items())
        items.sort()
        self.assertEqual(items,
                         [('eth_type', 0x0800), ('ipv4_src', '1.2.3.4')])

    def test_pktview_from_ofctl(self):
        data = dict(dl_type=0x0800, nw_proto=6, tp_dst=80)
        pkt = pktview_from_ofctl(data, validate=True)
        self.assertEqual(pkt, {'ip_proto': 6, 'eth_type': 2048, 'tcp_dst': 80})

        data = dict(dl_dst='ab:cd:00:00:00:00/ff:ff:00:00:00:00')
        pkt = pktview_from_ofctl(data, validate=True)
        self.assertEqual(pkt,
                         {'eth_dst': 'ab:cd:00:00:00:00/ff:ff:00:00:00:00'})
        items = pktview_to_list(pkt)
        self.assertEqual(items, [{
            'value': 'ab:cd:00:00:00:00',
            'field': 'ETH_DST',
            'mask': 'ff:ff:00:00:00:00'
        }])

        data = dict(tcp_dst='80/0xff')
        pkt = pktview_from_ofctl(data, validate=True)
        self.assertEqual(pkt, {'tcp_dst': '80/0xff'})
        items = pktview_to_list(pkt)
        self.assertEqual(items, [{
            'value': '80',
            'field': 'TCP_DST',
            'mask': '0xff'
        }])

    def test_convert_slash_notation(self):
        self.assertEqual(
            convert_slash_notation('ETH_SRC',
                                   '0e:dc:00:00:00:00/ff:ff:00:00:00:00'),
            ('0e:dc:00:00:00:00', 'ff:ff:00:00:00:00'))
        self.assertEqual(
            convert_slash_notation('IPV4_SRC', '1.2.3.4/255.255.0.0'),
            (IPv4Address('1.2.3.4'), IPv4Address('255.255.0.0')))
        self.assertEqual(
            convert_slash_notation('IPV4_SRC', '1.2.3.4/16'),
            (IPv4Address('1.2.3.4'), IPv4Address('255.255.0.0')))
        self.assertEqual(
            convert_slash_notation('IPV6_SRC', '2001::1/ffff::'),
            (IPv6Address('2001::1'), IPv6Address('ffff::')))
        self.assertEqual(
            convert_slash_notation('IPV6_SRC', '2001::1/16'),
            (IPv6Address('2001::1'), IPv6Address('ffff::')))

        addr = IPv4Address('1.2.3.4')
        self.assertIs(convert_slash_notation('IPV4_SRC', addr), addr)

        with self.assertRaises(ValueError):
            convert_slash_notation('IPV4_SRC', '1.2.3.4/64')

    def test_slash_notation_nd_target(self):
        result = convert_slash_notation('IPV6_ND_TARGET', 'fc00::1:2/112')
        self.assertEqual(result,
                         (IPv6Address('fc00::1:2'),
                          IPv6Address('ffff:ffff:ffff:ffff:ffff:ffff:ffff:0')))

        data = dict(ipv6_nd_target='fc00::1:2/112')
        fields = pktview_to_list(data)
        self.assertEqual(fields, [{
            'field': 'IPV6_ND_TARGET',
            'value': IPv6Address('fc00::1:2'),
            'mask': IPv6Address('ffff:ffff:ffff:ffff:ffff:ffff:ffff:0')
        }])

        result = pktview_from_list(fields, slash_notation=True)
        self.assertEqual(result, {
            'ipv6_nd_target': 'fc00::1:2/ffff:ffff:ffff:ffff:ffff:ffff:ffff:0'
        })

    def test_slash_notation_tcp_dst(self):
        result = convert_slash_notation('TCP_DST', '1024/1024')
        self.assertEqual(result, ('1024', '1024'))

        data = dict(tcp_dst='1024/1024')
        fields = pktview_to_list(data)
        self.assertEqual(fields, [{
            'field': 'TCP_DST',
            'value': '1024',
            'mask': '1024'
        }])

        result = pktview_from_list(fields, slash_notation=True)
        self.assertEqual(result, {'tcp_dst': '1024/1024'})

    def test_slash_notation_nd_sll(self):
        result = convert_slash_notation('IPV6_ND_SLL',
                                        '01:02:03:04:05:06/ff:ff:ff:00:00:00')
        self.assertEqual(result, ('01:02:03:04:05:06', 'ff:ff:ff:00:00:00'))

    def test_description(self):
        pkt = make_pktview(ip_proto=6, eth_type=0x0800)
        self.assertEqual(pkt.get_description(), 'TCPv4')
        pkt = make_pktview(ip_proto=1, eth_type=0x0800)
        self.assertEqual(pkt.get_description(), 'ICMPv4')
        pkt = make_pktview(ip_proto=2, eth_type=0x0800)
        self.assertEqual(pkt.get_description(), 'IGMPv4')
        pkt = make_pktview(ip_proto=17, eth_type=0x0800)
        self.assertEqual(pkt.get_description(), 'UDPv4')
        pkt = make_pktview(ip_proto=76, eth_type=0x0800)
        self.assertEqual(pkt.get_description(), 'IPv4:76')
        pkt = make_pktview(ip_proto=6, eth_type=0x86dd)
        self.assertEqual(pkt.get_description(), 'TCPv6')
        pkt = make_pktview(ip_proto=58, eth_type=0x86dd)
        self.assertEqual(pkt.get_description(), 'ICMPv6')
        pkt = make_pktview(ip_proto=17, eth_type=0x86dd)
        self.assertEqual(pkt.get_description(), 'UDPv6')
        pkt = make_pktview(eth_type=25)
        self.assertEqual(pkt.get_description(), 'ETH:25')
        pkt = make_pktview(eth_type=0x0800)
        self.assertEqual(pkt.get_description(), 'IPv4:None')
        pkt = make_pktview(ip_proto=79, eth_type=0x86dd)
        self.assertEqual(pkt.get_description(), 'IPv6:79')
        pkt = make_pktview(eth_type=0x0806, arp_op=1)
        self.assertEqual(pkt.get_description(), 'ARP:REQ')
        pkt = make_pktview(eth_type=0x0806, arp_op=2)
        self.assertEqual(pkt.get_description(), 'ARP:REPLY')
        pkt = make_pktview(eth_type=0x0806, arp_op=3)
        self.assertEqual(pkt.get_description(), 'ARP:3')
        pkt = make_pktview(eth_type=0x88cc)
        self.assertEqual(pkt.get_description(), 'LLDP')


def _by_field(item):
    return item['field']



import unittest
import zof


class TestApi(unittest.TestCase):
    def test_zof(self):
        zof.set_apps([])

        app1 = zof.Application('app1', exception_fatal=True, precedence=12)
        self.assertEqual(app1.name, 'app1')

        app2 = zof.Application('app2', precedence=5001)
        self.assertEqual(app2.name, 'app2')

        # Do not allow two apps to have the same name.
        app3 = zof.Application('app3')
        with self.assertRaises(ValueError):
            zof.Application('app3')

        # Test that apps are sorted by precedence.
        self.assertEqual(zof.get_apps(), [app2, app3, app1])



import unittest
from zof.event import load_event
from zof.objectview import to_json


class LoadEventTestCase(unittest.TestCase):
    def test_empty(self):
        # Loading empty bytes is an error.
        event = load_event(b'')
        self.assertEqual('EXCEPTION', event['event'])
        self.assertEqual('Expecting value: line 1 column 1 (char 0)',
                         event['reason'])

        # Loading only white space is an error.
        event = load_event(b'  ')
        self.assertEqual('EXCEPTION', event['event'])
        self.assertEqual('Expecting value: line 1 column 3 (char 2)',
                         event['reason'])

    def test_empty_dict(self):
        # Test empty event as bytes.
        event = load_event(b'{}')
        self.assertIsInstance(event, dict)
        self.assertEqual(event, {})

    def test_empty_dict2(self):
        # Test empty event as unicode str.
        event = load_event('{}')
        self.assertIsInstance(event, dict)
        self.assertEqual(event, {})

    def test_empty_dict3(self):
        # Test empty event with extra whitespace.
        event = load_event(b' {} \n')
        self.assertIsInstance(event, dict)
        self.assertEqual(event, {})

    def test_empty_dict_multiple(self):
        # Test that multiple empty events fail.
        event = load_event(b'{}{}')
        self.assertEqual('EXCEPTION', event['event'])
        self.assertEqual('Extra data: line 1 column 3 (char 2)',
                         event['reason'])

    def test_truncated_input(self):
        # Test that truncated input is detected.
        event = load_event(b'{ ')
        self.assertEqual('EXCEPTION', event['event'])
        self.assertEqual(
            'Expecting property name enclosed in double quotes: line 1 column 3 (char 2)',
            event['reason'])

    def test_other_value(self):
        # Test that non-mapping objects still work.
        event = load_event(b'[1, 2, 3]')
        self.assertIsInstance(event, list)
        self.assertEqual(event, [1, 2, 3])

    def test_utf8(self):
        # Test that Unicode/UTF-8 works.
        data = load_event(
            b'{"x":"\xe2\x82\xac\xf0\x90\x8c\x82\xf4\x8f\xbf\xbd"}')
        self.assertIsInstance(data, dict)
        self.assertEqual(data, {'x': '\u20AC\U00010302\U0010fffd'})

    def test_event_data(self):
        # Test that `data` is NOT converted from hexadecimal to binary.
        event = load_event(b'{"data":"deadbeef"}')
        self.assertEqual(event, {'data': 'deadbeef'})
        # Non-hex data should raise an exception. (FIXME- old)
        #event = load_event(b'{"data":"nonhex"}')
        #self.assertEqual('EXCEPTION', event['event'])
        #self.assertTrue(
        #    event['reason'].startswith('non-hexadecimal number found'))

    def test_event_pkt(self):
        # Test that `_pkt` lists are converted to dictionaries.
        #event = load_event(
        #    b'{"data": "ff", "_pkt": [{"field":"a", "value":1}]}')
        #self.assertEqual(event['pkt'], {'a': 1})
        #self.assertFalse("_pkt" in event)
        # _pkt requires the `data` value, otherwise it's not translated
        event = load_event(b'{"_pkt": [{"field":"a", "value":1}]}')
        self.assertFalse("pkt" in event)
        self.assertTrue("_pkt" in event)

    def test_event_payload(self):
        event = load_event(
            b'{"foo": 27, "data": "ff07", "_pkt": [{"field":"a", "value":9}, {"field": "X_PKT_POS", "value":1}]}'
        )
        self.assertEqual(event['foo'], 27)
        self.assertEqual(event['data'], 'ff07')
        #self.assertEqual(event['pkt'], {'a':9, 'x_pkt_pos': 1, 'payload': b'\x07'})
        #self.assertEqual(event['pkt']['payload'], b'\x07')

    def test_to_json(self):
        # Test that to_json also works.
        event = load_event(b'{"foo":"bar"}')
        s = to_json(event)
        self.assertEqual(s, '{"foo":"bar"}')

    def test_load_event_time(self):
        import timeit
        result = timeit.timeit(
            """load_event(b'{"foo": 27, "data": "ff07", "_pkt": [{"field":"a", "value":9}, {"field": "X_PKT_POS", "value":1}]}')""",
            setup='from zof.event import load_event',
            number=10000)
        print('[test_load_event_time=%r]' % result)



import unittest
import asyncio
import zof
from zof.controller import Controller


class SimulatorTester(zof.Application):
    def __init__(self, sim_app):
        super().__init__('test_simulator')
        self.count = 0
        self.sim_app = sim_app

        # Register instance methods by applying decorators.
        self.message('channel_up')(self.channel_up)
        self.message('channel_down')(self.channel_down)
        self.message('features_reply')(self.features_reply)
        self.message(any)(self.other_message)

    def channel_up(self, event):
        self.logger.info('channel_up')
        self.count += 1
        if self.count == self.sim_app.args.sim_count:
            self.logger.info('all channels up')
            zof.post_event({'event': 'EXIT', 'exit_status': 0})

    def channel_down(self, event):
        self.logger.info('channel_down')

    def features_reply(self, event):
        pass

    def other_message(self, event):
        self.logger.info(event)


class SimulatorTestCase(unittest.TestCase):
    def test_simulator(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        from zof.demo.simulator import APP as sim
        from zof.service.datapath import APP as dpa
        tester = SimulatorTester(sim)

        zof.set_apps([dpa, sim, tester])
        parser = zof.common_args(under_test=True)
        args = parser.parse_args([
            '--sim-count=50', '--sim-timeout=5',
            '--sim-endpoint=127.0.0.1:16653', '--listen-endpoints=16653'
        ])
        exit_status = zof.run(args=args)
        self.assertEqual(exit_status, 0)

        loop.close()



import unittest
import asyncio
from zof.protocol import Protocol
from zof.connection import Connection
from .asynctestcase import AsyncTestCase


class MockController:
    def __init__(self):
        self.events = []

    def post_event(self, event):
        self.events.append(event)


class ProtocolTestCase(unittest.TestCase):
    def test_data_received(self):
        controller = MockController()
        proto = Protocol(controller.post_event)
        proto.pipe_data_received(None, b'1\x002\x00')
        self.assertEqual(controller.events, [1, 2])
        self.assertEqual(proto.buf, b'')
        proto.pipe_data_received(None, b'')
        proto.pipe_data_received(None, b'3')
        proto.pipe_data_received(None, b'3')
        self.assertEqual(controller.events, [1, 2])
        self.assertEqual(proto.buf, b'33')
        proto.pipe_data_received(None, b'\x00')
        self.assertEqual(controller.events, [1, 2, 33])
        self.assertEqual(proto.buf, b'')
        proto.pipe_data_received(None, b'\x00\x00')
        self.assertEqual(controller.events, [1, 2, 33])
        self.assertEqual(proto.buf, b'')
        proto.pipe_data_received(None, b'"y')
        self.assertEqual(controller.events, [1, 2, 33])
        self.assertEqual(proto.buf, b'"y')
        proto.pipe_data_received(None, b'x"\x00')
        self.assertEqual(controller.events, [1, 2, 33, 'yx'])
        self.assertEqual(proto.buf, b'')


# Pass these args when launching oftr.
OFTR_ARGS = ''  # '--trace=rpc,msg --loglevel=debug'


class ProtocolConnectionTestCase(AsyncTestCase):
    async def setUp(self):
        oftr_options = {'args': OFTR_ARGS}
        self.conn = Connection(oftr_options=oftr_options)
        self.controller = MockController()
        await self.conn.connect(self.controller.post_event)
        assert self.conn.pid > 0

    async def tearDown(self):
        self.conn.close(True)
        return_code = await self.conn.disconnect()
        if return_code:
            raise Exception('oftr exited with return code %d' % return_code)

    async def test_rpc(self):
        # If we write a valid JSON-RPC method, we should get a result back.
        msg = b'{"id":1234,"method":"OFP.DESCRIPTION"}'
        self.conn.write(msg)
        for _ in range(10):
            await asyncio.sleep(0.01)
            if self.controller.events:
                break
        result = self.controller.events[0]
        self.assertEqual(result['id'], 1234)
        self.assertEqual(result['result']['api_version'], '0.9')



import unittest
from zof.endpoint import Endpoint


class EndpointTestCase(unittest.TestCase):
    def test_ipv4(self):
        endpt = Endpoint('127.0.0.1:80')
        self.assertEqual(str(endpt), '127.0.0.1:80')
        self.assertEqual(endpt.host, '127.0.0.1')
        self.assertEqual(endpt.port, 80)

    def test_ipv6(self):
        endpt = Endpoint('[::1]:81')
        self.assertEqual(str(endpt), '[::1]:81')
        self.assertEqual(endpt.host, '::1')
        self.assertEqual(endpt.port, 81)

    def test_port_only(self):
        endpt = Endpoint('82')
        self.assertEqual(str(endpt), ':82')
        self.assertEqual(endpt.host, '')
        self.assertEqual(endpt.port, 82)

        endpt = Endpoint(':82')
        self.assertEqual(str(endpt), ':82')
        self.assertEqual(endpt.host, '')
        self.assertEqual(endpt.port, 82)

        endpt = Endpoint('[]:82')
        self.assertEqual(str(endpt), ':82')
        self.assertEqual(endpt.host, '')
        self.assertEqual(endpt.port, 82)

    def test_invalid(self):
        subtests = [' :101', ':102 ']
        for subtest in subtests:
            with self.subTest(subtest=subtest):
                with self.assertRaises(ValueError):
                    Endpoint(subtest)



import zof

APP = zof.Application('cbenchreply')

FLOW_MOD = zof.compile('''
type: FLOW_MOD
flags: [NO_ALERT]
msg:
  table_id: 0
  command: ADD
  match:
   - field: ETH_DST
     value: $eth_dst
''')


@APP.message('packet_in')
def packet_in(event):
    FLOW_MOD.send(eth_dst=event['msg']['pkt'].eth_dst)


if __name__ == '__main__':
    zof.run()



from .asynctestcase import AsyncTestCase
from zof.http import HttpServer, HttpClient, ClientResponseError

import logging
from zof.logging import init_logging

init_logging('info')


class HttpTestCase(AsyncTestCase):
    async def test_http_server(self):
        logger = logging.getLogger('zof.test_http')
        web = HttpServer(logger=logger)

        @web.get_text('/')
        @web.get_text('/abc')
        async def _get_root():
            return 'Blah\n\u1E03lah'

        @web.get_text('/test/text')
        async def _get_text():
            return 'test text1'

        @web.get_json('/test/{var1}?{var2}')
        async def _get(var1, var2):
            return {'var1': var1, 'var2': var2}

        @web.post_json('/test/{var1}?{var2}&{var3}')
        async def _post(var1, var2, var3, post_data):
            return {
                'var1': var1,
                'var2': var2,
                'var3': var3,
                'post_data': post_data
            }

        await web.start('127.0.0.1:9010')

        client = HttpClient()
        await client.start()

        data = await client.get_text('http://127.0.0.1:9010/')
        self.assertEqual(data, 'Blah\n\u1E03lah')

        data = await client.get_text('http://127.0.0.1:9010/abc')
        self.assertEqual(data, 'Blah\n\u1E03lah')

        data = await client.get_json('http://127.0.0.1:9010/test/foo?var2=bar')
        self.assertEqual(data, {'var1': 'foo', 'var2': 'bar'})

        data = await client.get_json('http://127.0.0.1:9010/test/xyz')
        self.assertEqual(data, {'var1': 'xyz', 'var2': None})

        data = await client.post_json(
            'http://127.0.0.1:9010/test/a?var2=3&var3=4', post_data={'x': 1})
        self.assertEqual(data, {
            'var1': 'a',
            'var2': '3',
            'var3': '4',
            'post_data': {
                'x': 1
            }
        })

        data = await client.get_text('http://127.0.0.1:9010/test/text')
        self.assertEqual(data, 'test text1')

        # "/test" fails because {var1} is required.
        with self.assertRaises(ClientResponseError):
            await client.get_json('http://127.0.0.1:9010/test')

        # "/test/a/" doesn't match because of the / at the end.
        with self.assertRaises(ClientResponseError):
            await client.post_json(
                'http://127.0.0.1:9010/test/a/', post_data={})

        await client.stop()
        await web.stop()



import unittest
from zof.ofctl import convert_from_ofctl


class TestOfctl(unittest.TestCase):
    def test_tcpv4(self):
        ofctl = {
            'dl_src': '00:00:00:00:00:01',
            'dl_dst': '00:00:00:00:00:02',
            'dl_type': 0x0800,
            'nw_src': '192.168.1.1',
            'nw_dst': '192.168.1.2',
            'nw_proto': 6,
            'tp_src': 1001,
            'tp_dst': 1002,
            'other_field': 'other_value'
        }
        result = convert_from_ofctl(ofctl)
        self.assertEqual(result, {
            'eth_src': '00:00:00:00:00:01',
            'eth_dst': '00:00:00:00:00:02',
            'eth_type': 0x0800,
            'ipv4_src': '192.168.1.1',
            'ipv4_dst': '192.168.1.2',
            'ip_proto': 6,
            'tcp_src': 1001,
            'tcp_dst': 1002,
            'other_field': 'other_value'
        })

    def test_nw_src_dup(self):
        # Test duplicate field (nw_src = ipv4_src)
        ofctl = {
            'nw_src': '192.168.1.1',
            'ipv4_src': '192.168.1.2',
        }
        with self.assertRaisesRegex(ValueError, 'Duplicate ofctl field'):
            convert_from_ofctl(ofctl)

    def test_dl_vlan(self):
        ofctl = {
            'dl_src': '00:00:00:00:00:01',
            'dl_dst': '00:00:00:00:00:02',
            'dl_vlan': '0x123',
            'dl_type': 0x0800
        }
        result = convert_from_ofctl(ofctl, validate=True)
        self.assertEqual(result, {
            'eth_src': '00:00:00:00:00:01',
            'eth_dst': '00:00:00:00:00:02',
            'eth_type': 0x0800,
            'vlan_vid': 0x1123
        })

    def test_vlan_zero(self):
        ofctl = {'dl_vlan': '0x0'}
        result = convert_from_ofctl(ofctl)
        self.assertEqual(result, {'vlan_vid': 0})

        ofctl = {'dl_vlan': 0x64}
        result = convert_from_ofctl(ofctl)
        self.assertEqual(result, {'vlan_vid': 0x1064})

    def test_tp_src(self):
        # If ip_proto is not specified, default to 'tcp'
        ofctl = {'tp_src': 80, 'tp_dst': 81}
        result = convert_from_ofctl(ofctl)
        self.assertEqual(result, {'tcp_src': 80, 'tcp_dst': 81})

    def test_tp_fail(self):
        # If ip_proto is not supported, raise ValueError.
        ofctl = {'nw_proto': 2, 'tp_src': 80, 'tp_dst': 81}
        with self.assertRaises(ValueError):
            convert_from_ofctl(ofctl)

    def test_tp_dst(self):
        # UDP
        ofctl = {'dl_type': 0x0800, 'nw_proto': 17, 'tp_dst': 53}
        result = convert_from_ofctl(ofctl)
        self.assertEqual(result,
                         {'eth_type': 0x0800,
                          'ip_proto': 17,
                          'udp_dst': 53})

        ofctl['nw_proto'] = 6  # TCP
        result = convert_from_ofctl(ofctl)
        self.assertEqual(result,
                         {'eth_type': 0x0800,
                          'ip_proto': 6,
                          'tcp_dst': 53})

        ofctl['nw_proto'] = 132  # SCTP
        result = convert_from_ofctl(ofctl)
        self.assertEqual(
            result, {'eth_type': 0x0800,
                     'ip_proto': 132,
                     'sctp_dst': 53})

    def test_empty(self):
        result = convert_from_ofctl(None)
        self.assertEqual(result, None)

        result = convert_from_ofctl({})
        self.assertEqual(result, {})

    def test_arp(self):
        ofctl = {
            'dl_src': '00:00:00:00:00:01',
            'dl_dst': '00:00:00:00:00:02',
            'dl_type': 0x0806,
            'arp_op': 1,
            'nw_src': '192.168.1.1',
            'nw_dst': '192.168.1.2'
        }
        result = convert_from_ofctl(ofctl)
        self.assertEqual(result, {
            'eth_src': '00:00:00:00:00:01',
            'eth_dst': '00:00:00:00:00:02',
            'eth_type': 0x0806,
            'arp_op': 1,
            'arp_spa': '192.168.1.1',
            'arp_tpa': '192.168.1.2',
        })

    def test_validate_invalid(self):
        ofctl = {
            'dl_src': '00:00:00:00:00:01',
            'dl_dst': '00:00:00:00:00:02',
            'dl_type': 0x0800,
            'nw_src': '192.168.1.1',
            'nw_dst': '192.168.1.2',
            'nw_proto': 6,
            'tp_src': 1001,
            'tp_dst': 1002,
            'other_field': 'other_value'
        }
        with self.assertRaisesRegex(ValueError, 'other_field'):
            convert_from_ofctl(ofctl, validate=True)

        ofctl = {'udp_src': 'x'}
        with self.assertRaisesRegex(ValueError, 'udp_src: x'):
            convert_from_ofctl(ofctl, validate=True)

    def test_validate_masked(self):
        ofctl = {'udp_src': '123/0xFF'}
        result = convert_from_ofctl(ofctl, validate=True)
        self.assertEqual(result, {'udp_src': '123/0xFF'})

        ofctl = {'udp_src': '123/70/12'}
        with self.assertRaisesRegex(ValueError, 'udp_src: 123/70/12'):
            convert_from_ofctl(ofctl, validate=True)



from timeit import default_timer as timer
import zof
from zof.controller import Controller

APP = zof.Application('oftr_bench')
CONTROLLER = Controller.singleton()
N = 10000


@APP.event('start')
async def start(event):
    for i in range(5):
        start_time = timer()
        for _ in range(N):
            await CONTROLLER.rpc_call('OFP.DESCRIPTION')
        end_time = timer()
        elapsed = end_time - start_time
        APP.logger.info('Elapsed %r', elapsed)
    zof.post_event({'event': 'EXIT'})


if __name__ == '__main__':
    zof.run()
